import mimetypes
import os
import tempfile
import zipfile
from io import StringIO
from urllib.parse import urlparse

import weasyprint
from django.core.files.storage import default_storage
from rest_framework import renderers
from terra_opp.helpers import CustomCsvBuilder


class CSVRenderer(renderers.BaseRenderer):
    media_type = "text/csv"
    format = "csv"

    def render(self, data, media_type=None, renderer_context=None):
        csvbuilder = CustomCsvBuilder(data)
        csv_file = StringIO()

        csvbuilder.create_csv(csv_file)

        return csv_file.read()


def django_url_fetcher(url, *args, **kwargs):
    """Helper from django-weasyprint"""
    # load file:// paths directly from disk
    if url.startswith("file:"):
        mime_type, encoding = mimetypes.guess_type(url)
        parsed_url = urlparse(url)

        data = {
            "mime_type": mime_type,
            "encoding": encoding,
            "filename": parsed_url.netloc,
        }
        # try to find in media storage
        if default_storage.exists(parsed_url.netloc):
            data["file_obj"] = default_storage.open(parsed_url.netloc)
            return data

    # fall back to weasyprint default fetcher
    return weasyprint.default_url_fetcher(url, *args, **kwargs)


def write_pdf(request, html):
    """
    Function used to write pdf from html
    :param request: Django request
    :param html: HTML string
    :return: pdf as bytes
    """
    base_url = request.build_absolute_uri("/")

    return weasyprint.HTML(
        string=html,
        base_url=base_url,
        url_fetcher=django_url_fetcher,
    ).write_pdf()


class PdfRenderer(renderers.TemplateHTMLRenderer):
    media_type = "application/pdf"
    format = "pdf"

    def render(self, data, accepted_media_type=None, renderer_context=None):
        """Returns the rendered pdf"""
        html = super().render(
            data,
            accepted_media_type=accepted_media_type,
            renderer_context=renderer_context,
        )
        return write_pdf(renderer_context["request"], html)


class ZipRenderer(renderers.JSONRenderer):
    media_type = "application/zip"
    format = "zip"

    def render(self, data, media_type=None, renderer_context=None):
        if not isinstance(data, list):
            return data

        if renderer_context["request"].method != "OPTIONS":
            with tempfile.SpooledTemporaryFile() as tmp:
                with zipfile.ZipFile(tmp, "w", zipfile.ZIP_DEFLATED) as archive:
                    for file in data:
                        archive.writestr(
                            os.path.basename(file.name),
                            file.read(),
                        )
                tmp.seek(0)
                return tmp.read()
        return super(ZipRenderer, self).render(data, media_type, renderer_context)



# Generated by Django 2.2.18 on 2021-03-25 10:28

from django.db import migrations
from collections import defaultdict

state_map = defaultdict(
    lambda: "draft",
    {
        300: "accepted",
        200: "submited",
        -1: "refused",
    },
)

old_state_map = defaultdict(
    lambda: 100,
    {
        "accepted": 300,
        "submited": 200,
        "refused": -1,
    },
)


def migrate_picture_states(app, schema_editor):
    Picture = app.get_model("terra_opp", "Picture")
    for picture in Picture.objects.all():
        picture.state = state_map[picture.old_state]
        picture.save()


def reverse_migrate_picture_states(app, schema_editor):
    Picture = app.get_model("terra_opp", "Picture")
    for picture in Picture.objects.all():
        picture.old_state = old_state_map[picture.state]
        picture.save()


class Migration(migrations.Migration):

    dependencies = [
        ("terra_opp", "0007_auto_20210325_1022"),
    ]

    operations = [
        migrations.RunPython(
            migrate_picture_states, reverse_code=reverse_migrate_picture_states
        ),
        migrations.RemoveField(
            model_name="picture",
            name="old_state",
        ),
    ]



from typing import Optional

from django.conf import settings
from django.db import transaction
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from geostore import GeometryTypes
from geostore.models import Feature, Layer
from rest_framework import serializers, fields
from rest_framework_gis.fields import GeometryField
from datastore.models import RelatedDocument
from datastore.serializers import RelatedDocumentUrlSerializer
from datastore.fields import FileBase64UrlField
from versatileimagefield.serializers import VersatileImageFieldSerializer

from .models import Campaign, City, Picture, Theme, Viewpoint

UserModel = get_user_model()


class PermissiveImageFieldSerializer(VersatileImageFieldSerializer):
    def get_attribute(self, instance):
        try:
            return super().get_attribute(instance)
        except (AttributeError, ObjectDoesNotExist):
            # Will silence any NoneType or failing query on attribute
            return None


class CitySerializer(serializers.ModelSerializer):
    class Meta:
        model = City
        fields = "__all__"


class ThemeSerializer(serializers.ModelSerializer):
    class Meta:
        model = Theme
        fields = "__all__"


class LabelSlugRelatedField(serializers.SlugRelatedField):
    _model = None

    def to_internal_value(self, data):
        self._model.objects.get_or_create(label=data, defaults={"label": data})
        return super().to_internal_value(data)


class CityLabelSlugRelatedField(serializers.SlugRelatedField):
    _model = City

    def to_internal_value(self, data):
        data = data.capitalize()
        self._model.objects.get_or_create(label=data, defaults={"label": data})
        return super().to_internal_value(data)


class ThemeLabelSlugRelatedField(LabelSlugRelatedField):
    _model = Theme


class SimpleViewpointSerializer(serializers.ModelSerializer):
    picture = VersatileImageFieldSerializer("terra_opp")
    point = GeometryField(source="point.geom")
    city = CityLabelSlugRelatedField(
        slug_field="label",
        queryset=City.objects.all(),
        required=False,
        allow_null=False,
    )

    class Meta:
        model = Viewpoint
        fields = (
            "id",
            "label",
            "picture",
            "point",
            "active",
            "city",
        )


class SimpleAuthenticatedViewpointSerializer(SimpleViewpointSerializer):
    city = CityLabelSlugRelatedField(
        slug_field="label",
        queryset=City.objects.all(),
        required=False,
        allow_null=False,
    )
    themes = ThemeLabelSlugRelatedField(
        slug_field="label", many=True, queryset=Theme.objects.all(), required=False
    )
    last_accepted_picture_date = serializers.DateTimeField(read_only=True)

    class Meta:
        model = Viewpoint
        fields = (
            "id",
            "label",
            "picture",
            "point",
            "properties",
            "city",
            "themes",
            "active",
            "last_accepted_picture_date",
        )


class PhotographSerializer(serializers.ModelSerializer):
    class Meta:
        model = UserModel
        fields = ("uuid", "email", "properties")
        read_only_fields = (
            "uuid",
            UserModel.USERNAME_FIELD,
        )


class PictureSerializer(serializers.ModelSerializer):
    owner_id = serializers.PrimaryKeyRelatedField(
        source="owner",
        queryset=UserModel.objects.all(),
        required=False,
        many=False,
    )
    owner = PhotographSerializer(read_only=True)
    file = VersatileImageFieldSerializer("terra_opp")
    identifier = serializers.CharField(read_only=True)

    class Meta:
        model = Picture
        fields = "__all__"


class SimplePictureSerializer(PictureSerializer):
    class Meta:
        model = Picture
        fields = ("id", "date", "file", "owner", "properties", "state", "identifier")


class CampaignPictureSerializer(PictureSerializer):
    class Meta:
        model = Picture
        fields = ("id", "date", "state", "viewpoint")


class CustomRelatedDocumentUrlSerializer(RelatedDocumentUrlSerializer):
    document = FileBase64UrlField(required=False)


class ViewpointSerializerWithPicture(serializers.ModelSerializer):
    picture_ids = serializers.PrimaryKeyRelatedField(
        source="pictures",
        queryset=Picture.objects.all(),
        required=False,
        many=True,
    )
    pictures = SimplePictureSerializer(many=True, read_only=True)
    last_accepted_picture_date = serializers.DateTimeField(read_only=True)
    related = CustomRelatedDocumentUrlSerializer(many=True, required=False)
    point = GeometryField(source="point.geom")
    city = CityLabelSlugRelatedField(
        slug_field="label",
        queryset=City.objects.all(),
        required=False,
        allow_null=False,
    )
    themes = ThemeLabelSlugRelatedField(
        slug_field="label", many=True, queryset=Theme.objects.all(), required=False
    )

    class Meta:
        model = Viewpoint
        fields = (
            "id",
            "label",
            "properties",
            "point",
            "picture_ids",
            "pictures",
            "related",
            "city",
            "themes",
            "active",
            "last_accepted_picture_date",
        )

    @transaction.atomic
    def create(self, validated_data):
        related_docs = validated_data.pop("related", None)
        point_data = validated_data.pop("point", None)

        # Get or create layer
        layer, created = Layer.objects.get_or_create(
            pk=settings.TROPP_OBSERVATORY_LAYER_PK,
            defaults={
                "geom_type": GeometryTypes.Point,
                "id": settings.TROPP_OBSERVATORY_LAYER_PK,
            },
        )

        # Create feature
        feature = Feature.objects.create(
            geom=point_data.get("geom"),
            layer_id=settings.TROPP_OBSERVATORY_LAYER_PK,
            properties={},
        )
        validated_data.setdefault("point", feature)

        # Handle themes
        themes_labels = validated_data.pop("themes", None)
        if themes_labels:
            theme_list = []
            for theme_label in themes_labels:
                theme, created = Theme.objects.get_or_create(
                    label=theme_label,
                    defaults={
                        "label": theme_label,
                    },
                )
                theme_list.append(theme)
            validated_data.setdefault("themes", theme_list)

        instance = super().create(validated_data)

        # Handle related docs
        self.handle_related_documents(instance, related_docs)

        return instance

    @transaction.atomic
    def update(self, instance, validated_data):
        point_data = validated_data.pop("point", None)
        if point_data:
            feature = instance.point
            feature.geom = point_data.get("geom")
            feature.save()

        # Remove pictures no longer associated with viewpoint
        if "pictures" in validated_data:
            picture_ids = [p.pk for p in validated_data["pictures"]]
            instance.pictures.exclude(pk__in=picture_ids).delete()

        related_docs = validated_data.pop("related", None)
        self.handle_related_documents(instance, related_docs)

        return super().update(instance, validated_data)

    @staticmethod
    def handle_related_documents(instance: Viewpoint, related_docs: Optional[list]):
        if related_docs is not None:
            # Remove stale
            instance.related.exclude(key__in=[r["key"] for r in related_docs]).delete()
            for related in related_docs:
                file = None
                if "document" in related:
                    file = related["document"]
                    extension = file.content_type.split("/")[1]
                    file.name = f"{related['key']}.{extension}"

                try:
                    existing = instance.related.get(key=related["key"])
                    # Update file if sent
                    if file:
                        existing.document = file
                    # Update properties
                    existing.properties = related["properties"]
                    existing.save()
                except RelatedDocument.DoesNotExist:
                    RelatedDocument(**related, linked_object=instance).save()


class ViewpointLabelSerializer(serializers.ModelSerializer):
    class Meta:
        model = Viewpoint
        fields = ("id", "label")


# Writable serializer
class CampaignSerializer(serializers.ModelSerializer):
    start_date = fields.DateField(input_formats=["%Y-%m-%d"])
    owner = serializers.PrimaryKeyRelatedField(many=False, read_only=True)

    class Meta:
        model = Campaign
        fields = "__all__"


# ReadOnly serializer
class RoCampaignSerializer(CampaignSerializer):
    pictures = CampaignPictureSerializer(many=True, read_only=True)
    statistics = serializers.SerializerMethodField()

    # Format stats as dict
    def get_statistics(self, obj):
        return dict(
            total=obj.viewpoints_total,
            submited=obj.pictures_submited,
            accepted=obj.pictures_accepted,
            missing=obj.pictures_missing,
        )


# List serializer
class ListCampaignNestedSerializer(RoCampaignSerializer):
    class Meta(CampaignSerializer.Meta):
        model = Campaign
        fields = (
            "id",
            "label",
            "start_date",
            "assignee",
            "statistics",
            "state",
        )



from django.db.models.signals import post_save
from django.dispatch import receiver

from terra_opp.models import Picture


@receiver(post_save, sender=Picture)
def create_picture_identifier(sender, instance, **kwargs):
    if instance.state == sender.ACCEPTED and not instance.identifier:
        instance.identifier = instance.get_identifier()
        instance.save()



# Generated by Django 3.1.7 on 2021-04-09 09:04
from django.conf import settings
from django.db import migrations


def populate_picture_identifier(apps, schema_editor):
    obs_id = settings.TROPP_OBSERVATORY_ID or ""
    Picture = apps.get_model("terra_opp", "Picture")

    for picture in Picture.objects.all():
        viewpoint = picture.viewpoint
        pic_index = list(
            viewpoint.pictures.order_by("date").values_list("id", flat=True)
        ).index(picture.id)
        pic_index += 1  # list index start at 0 & we want the first to be 1
        picture.identifier = int(f"{obs_id}0{viewpoint.id:03}{pic_index:02}")
        picture.save()


class Migration(migrations.Migration):

    dependencies = [
        ("terra_opp", "0010_picture_identifier"),
    ]

    operations = [
        migrations.RunPython(populate_picture_identifier, migrations.RunPython.noop)
    ]



TROPP_URL_FETCHER = "terra_opp.renderers.django_url_fetcher"

TROPP_PICTURES_STATES_WORKFLOW = False

TROPP_VIEWPOINT_PROPERTIES_SET = {
    "pdf": {
        ("camera", "Appareil photo"),
    },
    "form": {},
    "filter": {},
}

TROPP_FEATURES_PROPERTIES_FROM_VIEWPOINT = ()

TROPP_SEARCHABLE_PROPERTIES = {
    "road": {
        "json_key": "voie",
        "type": "text",
    },
    "site": {
        "json_key": "site",
        "type": "text",
    },
}

# Create a geostore layer and provide its Primary Key here
TROPP_OBSERVATORY_LAYER_PK = None
TROPP_OBSERVATORY_ID = None

TERRA_APPLIANCE_SETTINGS = {}



# Generated by Django 3.2 on 2021-04-13 12:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("terra_opp", "0011_auto_20210412_1000"),
    ]

    operations = [
        migrations.RenameField(
            model_name="picture",
            old_name="identifier",
            new_name="old_identifier",
        ),
        migrations.AddField(
            model_name="picture",
            name="identifier",
            field=models.CharField(
                default="", max_length=10, verbose_name="Identifier"
            ),
        ),
    ]



from django.contrib import admin

from . import models


admin.site.register(models.Viewpoint)
admin.site.register(models.Campaign)
admin.site.register(models.Picture)
admin.site.register(models.City)
admin.site.register(models.Theme)



# Generated by Django 2.2.18 on 2021-02-24 12:43

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ("terra_opp", "0004_viewpoint_active"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="campaign",
            options={
                "ordering": ["-created_at"],
                "permissions": (("manage_campaigns", "Can manage all campaigns"),),
            },
        ),
        migrations.AddField(
            model_name="campaign",
            name="start_date",
            field=models.DateTimeField(
                default=django.utils.timezone.now, verbose_name="Start date"
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="campaign",
            name="label",
            field=models.CharField(max_length=250, verbose_name="Label"),
        ),
    ]



import base64
import os
from datetime import timedelta
from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.core.files.uploadedfile import SimpleUploadedFile, UploadedFile
from django.test import override_settings
from django.urls import reverse
from django.utils import timezone, dateparse
from rest_framework import status
from rest_framework.test import APITestCase

from terra_accounts.tests.factories import TerraUserFactory
from geostore.models import Feature
from geostore.tests.factories import FeatureFactory
from terra_opp.models import Picture, Viewpoint, City
from terra_opp.tests.factories import (
    CityFactory,
    PictureFactory,
    ThemeFactory,
    ViewpointFactory,
)
from terra_opp.tests.mixins import TestPermissionsMixin


@override_settings(TROPP_OBSERVATORY_ID=20)
class ViewpointCapitaliseTestCase(APITestCase, TestPermissionsMixin):
    @classmethod
    def setUpTestData(cls):
        cls.feature = FeatureFactory()
        cls.user = TerraUserFactory()

    def test_create_viewpoint(self):
        self.data_create = {
            "label": "Basic viewpoint created",
            "point": self.feature.geom.json,
            "city": "Marseille",
        }
        self.client.force_authenticate(user=self.user)
        self._set_permissions(["can_manage_viewpoints"])

        response = self.client.post(
            reverse("terra_opp:viewpoint-list"),
            self.data_create,
        )
        # Request is correctly constructed and viewpoint has been created
        self.assertEqual(status.HTTP_201_CREATED, response.status_code, response.json())
        self.assertEqual(response.json()["city"], "Marseille")
        self.assertTrue(City.objects.filter(label="Marseille").exists())

    def test_create_uncapitalized(self):
        self.data_create = {
            "label": "Basic viewpoint created",
            "point": self.feature.geom.json,
            "city": "marseille",
        }
        self.client.force_authenticate(user=self.user)
        self._set_permissions(["can_manage_viewpoints"])

        response = self.client.post(
            reverse("terra_opp:viewpoint-list"),
            self.data_create,
        )
        # Request is correctly constructed and viewpoint has been created
        self.assertEqual(status.HTTP_201_CREATED, response.status_code, response.json())
        self.assertEqual(response.json()["city"], "Marseille")
        self.assertTrue(City.objects.filter(label="Marseille").exists())
        self.assertFalse(City.objects.filter(label="marseille").exists())

    def test_already_in_db(self):
        City.objects.create(label="Marseille")
        self.data_create = {
            "label": "Basic viewpoint created",
            "point": self.feature.geom.json,
            "city": "marseille",
        }
        self.client.force_authenticate(user=self.user)
        self._set_permissions(["can_manage_viewpoints"])

        response = self.client.post(
            reverse("terra_opp:viewpoint-list"),
            self.data_create,
        )
        # Request is correctly constructed and viewpoint has been created
        self.assertEqual(status.HTTP_201_CREATED, response.status_code, response.json())
        self.assertEqual(response.json()["city"], "Marseille")
        self.assertTrue(City.objects.filter(label="Marseille").exists())
        self.assertFalse(City.objects.filter(label="marseille").exists())


@override_settings(TROPP_OBSERVATORY_ID=20)
class ViewpointTestCase(APITestCase, TestPermissionsMixin):
    @classmethod
    @override_settings(TROPP_PICTURES_STATES_WORKFLOW=True)
    def setUpTestData(cls):
        cls.feature = FeatureFactory()
        cls.user = TerraUserFactory()
        # Create viewpoint with draft picture attached to it
        cls.viewpoint = ViewpointFactory(label="Basic viewpoint")
        # Create viewpoint with accepted picture attached to it
        cls.viewpoint_with_accepted_picture = ViewpointFactory(
            label="Viewpoint with accepted picture",
            pictures__state="accepted",
            properties={"test_update": "ko"},
        )
        # Create viewpoints with no picture attached to it
        cls.viewpoint_without_picture = ViewpointFactory(
            label="Viewpoint without picture",
            pictures=None,
            properties={"test_update": "ko"},
        )

    def setUp(self):
        self.fp = open(
            os.path.join(os.path.dirname(__file__), "placeholder.jpg"),
            "rb",
        )

        self.data_create = {
            "label": "Basic viewpoint created",
            "point": self.feature.geom.json,
            "city": "Nantes",
        }

        self.data_create_with_picture = {
            "label": "Viewpoint created with picture",
            "point": self.feature.geom.json,
            "picture_ids": [
                picture.pk
                for picture in self.viewpoint_with_accepted_picture.pictures.all()
            ],
            "city": "Nantes",
        }

        self.data_create_with_themes = {
            "label": "Viewpoint created with themes",
            "point": self.feature.geom.json,
            "city": "Nantes",
            "themes": ["foo", "bar"],
        }
        self._clean_permissions()  # Don't forget that !

    def tearDown(self):
        self.fp.close()

    def _viewpoint_create(self):
        return self.client.post(
            reverse("terra_opp:viewpoint-list"),
            self.data_create,
        )

    def _viewpoint_create_with_picture(self):
        return self.client.post(
            reverse("terra_opp:viewpoint-list"),
            self.data_create_with_picture,
            format="multipart",
        )

    def test_viewpoint_get_list_anonymous(self):
        with self.assertNumQueries(6):
            data = self.client.get(reverse("terra_opp:viewpoint-list")).json()
        # List must contain all viewpoints WITHOUT those with no pictures
        # Pictures must also be ACCEPTED
        self.assertEqual(1, data.get("count"))

    def test_viewpoint_get_list_with_auth(self):
        # User is now authenticated
        self.client.force_authenticate(user=self.user)
        data = self.client.get(reverse("terra_opp:viewpoint-list")).json()
        # List must still contain ALL viewpoints even those with no
        # pictures and pictures with other states than ACCEPTED
        self.assertEqual(3, data.get("count"))

    def test_anonymous_access_without_accepted_picture(self):
        # User is not authenticated yet
        response = self.client.get(
            reverse(
                "terra_opp:viewpoint-detail",
                args=[self.viewpoint_without_picture.pk],
            )
        )
        # There is no picture on the viewpoint
        self.assertEqual(status.HTTP_404_NOT_FOUND, response.status_code)

    def test_viewpoint_get_with_auth(self):
        # User is now authenticated
        self.client.force_authenticate(user=self.user)
        response = self.client.get(
            reverse(
                "terra_opp:viewpoint-detail",
                args=[self.viewpoint_without_picture.pk],
            )
        )
        self.assertEqual(status.HTTP_200_OK, response.status_code)
        self.assertEqual(None, response.json()["last_accepted_picture_date"])

    def test_anonymous_options_request_returns_correct_search_filters(self):
        city1 = CityFactory(label="Montcuq")
        city2 = CityFactory(label="Rouperou-le-coquet")
        theme1 = ThemeFactory(label="Bar")
        theme2 = ThemeFactory(label="foo")
        ViewpointFactory(
            city=city1,
            themes=[theme1, theme2],
        )
        ViewpointFactory(
            city=city2,
            themes=[theme1],
        )
        data = self.client.get(reverse("terra_opp:viewpoint-filters")).json()
        self.assertEqual(data.get("cities"), ["Montcuq", "Rouperou-le-coquet"])
        self.assertEqual(data.get("themes"), ["Bar", "foo"])

    def test_authenticated_options_request_returns_all_search_filters(self):
        self.client.force_authenticate(user=self.user)
        data = self.client.get(reverse("terra_opp:viewpoint-filters")).json()
        self.assertIsNotNone(data.get("cities"))
        self.assertIsNotNone(data.get("themes"))

        # Even if we have 3 users, we only get those who have pictures
        self.assertEqual(3, get_user_model().objects.count())
        self.assertEqual(2, len(data.get("photographers")))

    def test_viewpoint_search_anonymous(self):
        # Simple viewpoint search feature
        data = self.client.get(
            reverse("terra_opp:viewpoint-list"),
            {"search": "accepted"},
        ).json()
        self.assertEqual(data.get("count"), 1)

    def test_viewpoint_search_with_auth(self):
        # Simple viewpoint search feature with auth
        self.client.force_authenticate(user=self.user)
        data = self.client.get(
            reverse("terra_opp:viewpoint-list"),
            {"search": "Basic"},
        ).json()
        self.assertEqual(data.get("count"), 1)

    def test_viewpoint_picture_filter_anonymous(self):
        data = self.client.get(
            reverse("terra_opp:viewpoint-list"),
            {
                "pictures__identifier": self.viewpoint_with_accepted_picture.pictures.first().identifier
            },
        ).json()
        self.assertEqual(data.get("count"), 1)

    def test_viewpoint_picture_filter_with_auth(self):
        self.client.force_authenticate(user=self.user)
        data = self.client.get(
            reverse("terra_opp:viewpoint-list"),
            {
                "pictures__identifier": self.viewpoint_with_accepted_picture.pictures.first().identifier
            },
        ).json()
        self.assertEqual(data.get("count"), 1)

    def test_viewpoint_photographer_filter_anonymous(self):
        picture = self.viewpoint_with_accepted_picture.pictures.first()
        data = self.client.get(
            reverse("terra_opp:viewpoint-list"),
            {"pictures__owner__uuid": picture.owner.uuid},
        ).json()
        self.assertEqual(data.get("count"), 1)

    def test_viewpoint_photographer_filter_with_auth(self):
        self.client.force_authenticate(user=self.user)
        picture = self.viewpoint_with_accepted_picture.pictures.first()
        data = self.client.get(
            reverse("terra_opp:viewpoint-list"),
            {"pictures__owner__uuid": picture.owner.uuid},
        ).json()
        self.assertEqual(data.get("count"), 1)

    def test_viewpoint_search_date(self):
        list_url = reverse("terra_opp:viewpoint-list")
        picture = self.viewpoint_with_accepted_picture.pictures.first()
        data = self.client.get(
            list_url, {"date_from": (picture.date - timedelta(days=1)).date()}
        ).json()
        self.assertEqual(data.get("count"), 1)
        data = self.client.get(
            list_url, {"date_from": (picture.date + timedelta(days=1)).date()}
        ).json()
        self.assertEqual(data.get("count"), 0)
        data = self.client.get(
            list_url, {"date_to": (picture.date + timedelta(days=1)).date()}
        ).json()
        self.assertEqual(data.get("count"), 1)
        data = self.client.get(
            list_url, {"date_to": (picture.date - timedelta(days=1)).date()}
        ).json()
        self.assertEqual(data.get("count"), 0)
        data = self.client.get(
            list_url,
            {
                "date_from": (picture.date - timedelta(days=1)).date(),
                "date_to": (picture.date + timedelta(days=1)).date(),
            },
        ).json()
        self.assertEqual(data.get("count"), 1)

        # Errors
        response = self.client.get(list_url, {"date_to": "haha"})
        self.assertEqual(status.HTTP_400_BAD_REQUEST, response.status_code)

    def test_viewpoint_search_json(self):
        list_url = reverse("terra_opp:viewpoint-list")
        ViewpointFactory(
            label="Viewpoint for search",
            pictures__state="accepted",
            properties={
                "voie": "coin d'en bas de la rue du bout",
                "site": "Carrière des petits violoncelles",
            },
        )
        self.assertEqual(self.client.get(list_url).json()["count"], 2)
        data = self.client.get(list_url, {"properties__voie": "rue"}).json()
        self.assertEqual(data.get("count"), 1)
        data = self.client.get(list_url, {"properties__site": "carrière"}).json()
        self.assertEqual(data.get("count"), 1)

    def test_viewpoint_search_city(self):
        list_url = reverse("terra_opp:viewpoint-list")
        city = CityFactory(label="Rouperou-le-coquet")
        city2 = CityFactory(label="New york")
        city3 = CityFactory(label="Los angeles")
        ViewpointFactory(
            label="Viewpoint for search",
            pictures__state="accepted",
            city=city,
        )
        ViewpointFactory(
            label="Not found",
            pictures__state="accepted",
            city=city2,
        )
        ViewpointFactory(
            label="Not found also",
            pictures__state="accepted",
            city=city3,
        )
        self.assertEqual(self.client.get(list_url).json()["count"], 4)
        data = self.client.get(list_url, {"city_id": city.id}).json()
        self.assertEqual(data.get("count"), 1)
        data = self.client.get(list_url, {"city": "Rouperou-le-coquet"}).json()
        self.assertEqual(data.get("count"), 1)

    def test_viewpoint_search_themes(self):
        list_url = reverse("terra_opp:viewpoint-list")
        theme_foo = ThemeFactory(label="foo")
        theme_bar = ThemeFactory(label="bar")
        theme_baz = ThemeFactory(label="baz")
        theme_not = ThemeFactory(label="not")
        vp = ViewpointFactory(
            label="Viewpoint for search",
            pictures__state="accepted",
            properties={
                "voie": "coin d'en bas de la rue du bout",
                "site": "Carrière des petits violoncelles",
            },
        )
        vp.themes.add(theme_foo, theme_bar, theme_baz)
        self.assertEqual(self.client.get(list_url).json()["count"], 2)
        data = self.client.get(list_url, {"themes_id": [theme_foo.id]}).json()
        self.assertEqual(data.get("count"), 1)
        data = self.client.get(list_url, {"themes": ["foo"]}).json()
        self.assertEqual(data.get("count"), 1)
        data = self.client.get(
            list_url, {"themes_id": [theme_bar.id, theme_foo.id]}
        ).json()
        self.assertEqual(data.get("count"), 1)
        data = self.client.get(
            list_url, {"themes_id": [theme_bar.id, theme_not.id]}
        ).json()
        self.assertEqual(data.get("count"), 0)

    def test_viewpoint_create_anonymous(self):
        response = self._viewpoint_create()
        # User is not authenticated
        self.assertIn(
            response.status_code,
            [status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN],
        )

    def test_viewpoint_create_with_auth(self):
        self.client.force_authenticate(user=self.user)
        response = self._viewpoint_create()
        # User doesn't have permission
        self.assertEqual(status.HTTP_403_FORBIDDEN, response.status_code)

    def test_viewpoint_create_with_auth_and_perms(self):
        self.client.force_authenticate(user=self.user)
        self._set_permissions(
            [
                "can_manage_viewpoints",
            ]
        )
        response = self._viewpoint_create()
        # Request is correctly constructed and viewpoint has been created
        self.assertEqual(status.HTTP_201_CREATED, response.status_code, response.json())

    def test_viewpoint_create_with_picture_anonymous(self):
        response = self._viewpoint_create_with_picture()
        # User is not authenticated
        self.assertIn(
            response.status_code,
            [status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN],
        )

    def test_viewpoint_create_with_picture_with_auth(self):
        self.client.force_authenticate(user=self.user)
        response = self._viewpoint_create_with_picture()
        # User doesn't have permission
        self.assertEqual(status.HTTP_403_FORBIDDEN, response.status_code)

    def test_viewpoint_create_with_picture_with_auth_and_perms(self):
        self.client.force_authenticate(user=self.user)
        self._set_permissions(
            [
                "can_manage_viewpoints",
            ]
        )
        response = self._viewpoint_create_with_picture()

        # Request is correctly constructed and viewpoint has been created
        self.assertEqual(status.HTTP_201_CREATED, response.status_code, response.json())
        viewpoint = Viewpoint.objects.get(label="Viewpoint created with picture")
        self.assertIn(
            "/2018-01-01_00-00-00",
            viewpoint.point.properties["viewpoint_picture"],
        )

    def test_viewpoint_create_with_themes_with_auth_and_perms(self):
        self.client.force_authenticate(user=self.user)
        self._set_permissions(
            [
                "can_manage_viewpoints",
            ]
        )
        response = self.client.post(
            reverse("terra_opp:viewpoint-list"),
            self.data_create_with_themes,
        )
        # Request is correctly constructed and viewpoint has been created
        self.assertEqual(status.HTTP_201_CREATED, response.status_code)
        self.assertListEqual(
            list(
                Viewpoint.objects.get(label="Viewpoint created with themes")
                .themes.all()
                .values_list("label", flat=True)
            ),
            ["foo", "bar"],
        )

    @patch("datastore.fields.FileBase64Field.to_internal_value")
    def test_viewpoint_create_with_related_docs(self, field):
        self.client.force_authenticate(user=self.user)
        self._set_permissions(
            [
                "can_manage_viewpoints",
            ]
        )
        self.fp.seek(0)
        document = (
            f"data:image/jpg;base64,"
            f'{(base64.b64encode(self.fp.read())).decode("utf-8")}'
        )
        field.return_value = UploadedFile(
            self.fp,
            content_type="image/jpeg",
        )
        response = self.client.post(
            reverse("terra_opp:viewpoint-list"),
            {
                "label": "Viewpoint created with picture",
                "point": self.feature.geom.json,
                "related": [
                    {
                        "key": "croquis",
                        "document": document,
                    }
                ],
                "city": "Nantes",
            },
            format="json",
        )
        # Request is correctly constructed and viewpoint has been created
        self.assertEqual(status.HTTP_201_CREATED, response.status_code)
        data = response.json()
        self.assertEqual(1, len(data["related"]))
        self.assertEqual("croquis", data["related"][0]["key"])

        # Update it
        response = self.client.patch(
            reverse("terra_opp:viewpoint-detail", args=[data["id"]]),
            {
                "related": [
                    {
                        "key": "emplacement",
                        "document": document,
                    }
                ]
            },
            format="json",
        )
        # Request is correctly constructed and viewpoint has been updated
        self.assertEqual(status.HTTP_200_OK, response.status_code)
        data = response.json()
        self.assertEqual(1, len(data["related"]))
        self.assertEqual("emplacement", data["related"][0]["key"])

    def _viewpoint_delete(self):
        return self.client.delete(
            reverse("terra_opp:viewpoint-detail", args=[self.viewpoint.pk])
        )

    def test_viewpoint_delete_anonymous(self):
        response = self._viewpoint_delete()
        # User is not authenticated
        self.assertIn(
            response.status_code,
            [status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN],
        )

    def test_viewpoint_delete_with_auth(self):
        self.client.force_authenticate(user=self.user)
        response = self._viewpoint_delete()
        # User doesn't have permission
        self.assertEqual(status.HTTP_403_FORBIDDEN, response.status_code)

    def test_viewpoint_delete_with_auth_and_perms(self):
        self.client.force_authenticate(user=self.user)
        self._set_permissions(
            [
                "can_manage_viewpoints",
            ]
        )
        response = self._viewpoint_delete()
        # User have permission
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)

    def _viewpoint_update(self):
        return self.client.patch(
            reverse(
                "terra_opp:viewpoint-detail",
                args=[self.viewpoint_with_accepted_picture.pk],
            ),
            {
                "label": "test",
                "properties": {"test_update": "ok"},
                "point": {"type": "Point", "coordinates": [0.0, 1.0]},
            },
            format="json",
        )

    def test_viewpoint_update_anonymous(self):
        response = self._viewpoint_update()
        # User is not authenticated
        self.assertIn(
            response.status_code,
            [status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN],
        )

    def test_viewpoint_update_with_auth(self):
        self.client.force_authenticate(user=self.user)
        response = self._viewpoint_update()
        # User is authenticated but doesn't have permission to update the
        # viewpoint.
        self.assertEqual(status.HTTP_403_FORBIDDEN, response.status_code)

    def test_viewpoint_update_with_auth_and_perms(self):
        self.client.force_authenticate(user=self.user)
        self._set_permissions(
            [
                "can_manage_viewpoints",
            ]
        )

        response = self._viewpoint_update()

        # User is authenticated and have permission to update the viewpoint.
        self.assertEqual(status.HTTP_200_OK, response.status_code)

        # Check if the viewpoint is correctly updated
        viewpoint = Viewpoint.objects.get(pk=self.viewpoint_with_accepted_picture.pk)
        self.assertEqual(response.data["label"], viewpoint.label)
        self.assertEqual(
            response.data["properties"]["test_update"],
            viewpoint.properties["test_update"],
        )

        # Check if the viewpoint's feature is correctly updated
        feature = Feature.objects.get(pk=self.viewpoint_with_accepted_picture.point.pk)
        self.assertEqual(response.data["label"], feature.properties["viewpoint_label"])
        self.assertEqual(
            self.viewpoint_with_accepted_picture.pk, feature.properties["viewpoint_id"]
        )
        self.assertEqual(
            response.data["point"]["coordinates"],
            [feature.geom.coords[0], feature.geom.coords[1]],
        )

    def test_add_picture_on_viewpoint_with_auth_and_perms(self):
        self.client.force_authenticate(user=self.user)
        self._set_permissions(
            [
                "can_manage_viewpoints",
            ]
        )

        # We create a more recent picture
        date = timezone.datetime(2019, 1, 1, tzinfo=timezone.utc)

        response = self.client.get(
            reverse(
                "terra_opp:viewpoint-detail",
                args=[
                    self.viewpoint_with_accepted_picture.pk,
                ],
            ),
        )
        # Before last accepted date should be in 2018
        self.assertEqual(
            timezone.datetime(2018, 1, 1, tzinfo=timezone.utc),
            dateparse.parse_datetime(response.json()["last_accepted_picture_date"]),
        )

        file = SimpleUploadedFile(
            name="test.jpg",
            content=open(
                "terra_opp/tests/placeholder.jpg",
                "rb",
            ).read(),
            content_type="image/jpeg",
        )
        picture = Picture.objects.create(
            viewpoint=self.viewpoint_with_accepted_picture,
            owner=self.user,
            date=date,
            file=file,
            state="accepted",
        )
        response = self.client.patch(
            reverse(
                "terra_opp:viewpoint-detail",
                args=[
                    self.viewpoint_with_accepted_picture.pk,
                ],
            ),
            {"picture_ids": [picture.id]},
        )

        self.assertEqual(status.HTTP_200_OK, response.status_code)

        # New last picture date now
        self.assertEqual(
            date,
            dateparse.parse_datetime(response.json()["last_accepted_picture_date"]),
        )

        viewpoint = Viewpoint.objects.get(pk=self.viewpoint_with_accepted_picture.pk)
        self.assertEqual(1, viewpoint.pictures.count())
        self.assertIn(
            f"viewpoint_{self.viewpoint_with_accepted_picture.pk}/2019-01-01_00-00-00",
            viewpoint.pictures.latest().file.name,
        )

        feature = Feature.objects.get(pk=self.viewpoint_with_accepted_picture.point.pk)
        # Check if the feature has been updated after patching
        self.assertIn(file.name.split(".")[0], feature.properties["viewpoint_picture"])

    def test_add_older_picture_on_viewpoint_with_auth_and_perms(self):
        self.client.force_authenticate(user=self.user)
        self._set_permissions(
            [
                "can_manage_viewpoints",
            ]
        )

        # We create an older picture
        date = timezone.datetime(1950, 1, 1, tzinfo=timezone.utc)
        file = SimpleUploadedFile(
            name="test_older.jpg",
            content=open(
                "terra_opp/tests/placeholder.jpg",
                "rb",
            ).read(),
            content_type="image/jpeg",
        )
        picture = Picture.objects.create(
            viewpoint=self.viewpoint_with_accepted_picture,
            owner=self.user,
            date=date,
            file=file,
            state="accepted",
        )
        picture_ids = [
            pic.id for pic in self.viewpoint_with_accepted_picture.pictures.all()
        ]
        picture_ids.append(picture.id)
        response = self.client.patch(
            reverse(
                "terra_opp:viewpoint-detail",
                args=[
                    self.viewpoint_with_accepted_picture.pk,
                ],
            ),
            {"picture_ids": picture_ids},
        )
        self.assertEqual(status.HTTP_200_OK, response.status_code)

        viewpoint = Viewpoint.objects.get(pk=self.viewpoint_with_accepted_picture.pk)
        self.assertEqual(2, viewpoint.pictures.count())
        self.assertNotIn(file.name.split(".")[0], viewpoint.pictures.latest().file.name)

        feature = Feature.objects.get(pk=self.viewpoint_with_accepted_picture.point.pk)
        # Check that the feature has not been updated after patching
        self.assertNotIn(
            file.name.split(".")[0], feature.properties["viewpoint_picture"]
        )

    def test_ordering_in_list_view(self):
        self.client.force_authenticate(user=self.user)
        data = self.client.get(reverse("terra_opp:viewpoint-list")).json()
        # Now test that viewpoints are ordered in chronological order
        first_viewpoint = Viewpoint.objects.get(id=data.get("results")[0]["id"])
        second_viewpoint = Viewpoint.objects.get(id=data.get("results")[1]["id"])
        self.assertTrue(first_viewpoint.created_at > second_viewpoint.created_at)

    def test_list_viewset_return_distinct_objects(self):
        # We add a more recent picture to the viewpoint
        date = timezone.datetime(2019, 1, 1, tzinfo=timezone.utc)
        file = SimpleUploadedFile(
            name="test.jpg",
            content=open(
                "terra_opp/tests/placeholder.jpg",
                "rb",
            ).read(),
            content_type="image/jpeg",
        )
        Picture.objects.create(
            viewpoint=self.viewpoint_with_accepted_picture,
            owner=self.user,
            date=date,
            file=file,
            state="accepted",
        )

        # Viewpoint should appears only once in the list
        data = self.client.get(reverse("terra_opp:viewpoint-list")).json()
        self.assertEqual(1, data.get("count"))

    def test_pdf_view_must_return_pdf_when_authenticated(self):
        self.client.force_authenticate(user=self.user)
        data = self.client.get(
            reverse(
                "terra_opp:viewpoint-pdf",
                args=[self.viewpoint.pk],
            )
        )
        self.assertEqual(status.HTTP_200_OK, data.status_code)
        self.assertIn("application/pdf", data["Content-Type"])

    def test_options_request_on_zip_pictures_must_return_200(self):
        data = self.client.options(
            reverse(
                "terra_opp:viewpoint-zip-pictures",
                args=[
                    self.viewpoint_with_accepted_picture.pk,
                ],
            )
        )
        self.assertEqual(status.HTTP_200_OK, data.status_code)
        self.assertIn("application/zip", data["Content-Type"])

    def test_get_request_on_zip_pictures_must_return_a_zip_with_accepted_pictures(self):
        data = self.client.get(
            reverse(
                "terra_opp:viewpoint-zip-pictures",
                args=[
                    self.viewpoint_with_accepted_picture.pk,
                ],
            )
        )
        self.assertEqual(status.HTTP_200_OK, data.status_code)
        self.assertIn("application/zip", data["Content-Type"])
        self.assertEqual(1, len(data.data))

    def test_add_picture_on_viewpoint_must_update_feature_properties(self):
        self.client.force_authenticate(user=self.user)
        self._set_permissions(
            [
                "can_manage_pictures",
            ]
        )

        # should be equal to 1
        picture_count = self.viewpoint_with_accepted_picture.pictures.count()

        file = SimpleUploadedFile(
            name="add_picture_test.jpg",
            content=open(
                "terra_opp/tests/placeholder.jpg",
                "rb",
            ).read(),
            content_type="image/jpeg",
        )
        data = {
            "viewpoint": self.viewpoint_with_accepted_picture.pk,
            "date": timezone.datetime(2020, 8, 19, tzinfo=timezone.utc),
            "file": file,
            "state": "accepted",
        }
        # this request must create a new picture on the viewpoint and update it's feature properties
        response = self.client.post(
            reverse("terra_opp:picture-list"),
            data,
            format="multipart",
        )
        self.assertEqual(status.HTTP_201_CREATED, response.status_code)
        self.assertEqual(
            self.viewpoint_with_accepted_picture.pictures.count(), picture_count + 1
        )
        # Check if the feature has been updated after patching, so update the object from the DB before
        self.viewpoint_with_accepted_picture.refresh_from_db()
        self.assertIn(
            f"viewpoint_{self.viewpoint_with_accepted_picture.pk}/2020-08-19_00-00-00",
            self.viewpoint_with_accepted_picture.point.properties["viewpoint_picture"],
        )

    def test_add_picture_on_viewpoint_with_no_picture_must_update_feature_properties(
        self,
    ):
        self.client.force_authenticate(user=self.user)
        self._set_permissions(
            [
                "can_manage_pictures",
            ]
        )

        # should be equal to 0
        picture_count = self.viewpoint_without_picture.pictures.count()

        file = SimpleUploadedFile(
            name="add_picture_test.jpg",
            content=open(
                "terra_opp/tests/placeholder.jpg",
                "rb",
            ).read(),
            content_type="image/jpeg",
        )
        data = {
            "viewpoint": self.viewpoint_without_picture.pk,
            "date": timezone.datetime(2020, 8, 19, tzinfo=timezone.utc),
            "file": file,
            "state": "accepted",
        }
        # this request must create a new picture on the viewpoint and update it's feature properties
        response = self.client.post(
            reverse("terra_opp:picture-list"),
            data,
            format="multipart",
        )
        self.assertEqual(status.HTTP_201_CREATED, response.status_code)
        self.assertEqual(
            self.viewpoint_without_picture.pictures.count(), picture_count + 1
        )
        # Check if the feature has been updated after patching, so update the object from the DB before
        self.viewpoint_without_picture.refresh_from_db()
        self.assertIn(
            f"viewpoint_{self.viewpoint_without_picture.pk}/2020-08-19_00-00-00",
            self.viewpoint_without_picture.point.properties["viewpoint_picture"],
        )

    def test_update_picture_on_viewpoint_must_update_feature_properties(self):
        self.client.force_authenticate(user=self.user)
        self._set_permissions(
            [
                "can_manage_pictures",
            ]
        )

        file = SimpleUploadedFile(
            name="another_placeholder.jpg",
            content=open(
                "terra_opp/tests/another_placeholder.jpg",
                "rb",
            ).read(),
            content_type="image/jpeg",
        )
        data = {
            "file": file,
        }
        # this request must create a new picture on the viewpoint and update it's feature properties
        response = self.client.patch(
            reverse(
                "terra_opp:picture-detail",
                args=[self.viewpoint_with_accepted_picture.pictures.latest().pk],
            ),
            data,
            format="multipart",
        )
        self.assertEqual(status.HTTP_200_OK, response.status_code)
        # Check if the feature has been updated after patching, so update the object from the DB before
        self.viewpoint_with_accepted_picture.refresh_from_db()
        self.assertIn(
            f"viewpoint_{self.viewpoint_with_accepted_picture.pk}/2018-01-01_00-00-00",
            self.viewpoint_with_accepted_picture.point.properties["viewpoint_picture"],
        )

    def test_delete_picture_on_viewpoint_must_update_feature_properties(self):
        self.client.force_authenticate(user=self.user)
        self._set_permissions(
            [
                "can_manage_pictures",
            ]
        )

        PictureFactory(
            viewpoint=self.viewpoint_with_accepted_picture,
            date=timezone.datetime(2020, 8, 20, tzinfo=timezone.utc),
        )
        picture_count = self.viewpoint_with_accepted_picture.pictures.count()
        before_last_picture = self.viewpoint_with_accepted_picture.pictures.all()[1]

        # this request must delete the picture on the viewpoint and update it's feature properties
        response = self.client.delete(
            reverse(
                "terra_opp:picture-detail",
                args=[self.viewpoint_with_accepted_picture.pictures.latest().pk],
            ),
        )
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)
        self.assertEqual(
            self.viewpoint_with_accepted_picture.pictures.count(), picture_count - 1
        )
        # Check if the feature has been updated after patching, so update the object from the DB before
        self.viewpoint_with_accepted_picture.refresh_from_db()
        self.assertIn(
            before_last_picture.file.name.split(".")[0],
            self.viewpoint_with_accepted_picture.point.properties["viewpoint_picture"],
        )

    def test_delete_the_only_available_picture_on_viewpoint_must_update_feature_properties(
        self,
    ):
        self.client.force_authenticate(user=self.user)
        self._set_permissions(
            [
                "can_manage_pictures",
            ]
        )

        # The viewpoint must only have a single available picture for this test
        self.assertEqual(1, self.viewpoint_with_accepted_picture.pictures.count())

        # This request must delete the only available picture on the viewpoint.
        response = self.client.delete(
            reverse(
                "terra_opp:picture-detail",
                args=[self.viewpoint_with_accepted_picture.pictures.latest().pk],
            ),
        )
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)
        self.assertEqual(0, self.viewpoint_with_accepted_picture.pictures.count())
        # Check if the feature has been updated after patching, so update the object from the DB before
        self.viewpoint_with_accepted_picture.refresh_from_db()
        self.assertFalse(
            "viewpoint_picture" in self.viewpoint_with_accepted_picture.point.properties
        )

    def test_only_active_viewpoint_retrieve(self):
        active_viewpoint = ViewpointFactory(
            label="Active viewpoint",
            active=True,
            pictures__state="accepted",
        )
        inactive_viewpoint = ViewpointFactory(
            label="Unactive viewpoint",
            active=False,
            pictures__state="accepted",
        )
        response = self.client.get(reverse("terra_opp:viewpoint-active"))
        self.assertEqual(status.HTTP_200_OK, response.status_code)

        active_viewpoint_count = (
            Viewpoint.objects.with_accepted_pictures().filter(active=True).count()
        )
        data = response.json()

        self.assertEqual(active_viewpoint_count, data["count"])
        self.assertNotIn(inactive_viewpoint.id, [d["id"] for d in data["results"]])
        self.assertIn(active_viewpoint.id, [d["id"] for d in data["results"]])



from rest_framework.routers import SimpleRouter

from . import views

app_name = "terra_opp"

router = SimpleRouter()

router.register(r"viewpoints", views.ViewpointViewSet, basename="viewpoint")
router.register(r"campaigns", views.CampaignViewSet, basename="campaign")
router.register(r"pictures", views.PictureViewSet, basename="picture")
router.register(r"cities", views.CityViewSet, basename="city")
router.register(r"themes", views.ThemeViewSet, basename="theme")

urlpatterns = router.urls



# Generated by Django 2.2.18 on 2021-04-06 15:13

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("terra_opp", "0008_migrate_state"),
    ]

    operations = [
        migrations.AddField(
            model_name="picture",
            name="campaign",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="pictures",
                to="terra_opp.Campaign",
            ),
        ),
    ]



from django.core.checks import Warning, Error, register
from django.conf import settings
from geostore import models
from django.db.utils import ProgrammingError


@register()
def check_dedicated_layer(app_configs, **kwargs):
    errors = []
    observatory_layer_pk = settings.TROPP_OBSERVATORY_LAYER_PK

    if not observatory_layer_pk:
        errors.append(
            Warning(
                "To correctly use OPP You should create a dedicated layer, set TROPP_OBSERVATORY_LAYER_PK and restart your instance as soon as possible.",
                hint="""
                    Create a dedicated point layer with ./manage.py create_observatory_layer and set
                    TROPP_OBSERVATORY_LAYER_PK with given PK.
                    ex: TROPP_OBSERVATORY_LAYER_PK=4
                    """,
                obj=None,
                id="terra_opp.E002",
            )
        )
    else:
        try:
            models.Layer.objects.get(id=observatory_layer_pk)
        except models.Layer.DoesNotExist:
            errors.append(
                Warning(
                    f"Layer with id={observatory_layer_pk} seems missing. Are you sure you've created a layer with this id?",
                    hint="""
                        Create a dedicated point layer with ./manage.py create_observatory_layer and set
                        TROPP_OBSERVATORY_LAYER_PK with given PK.
                        ex: TROPP_OBSERVATORY_LAYER_PK=4
                        """,
                    obj=None,
                    id="terra_opp.E003",
                )
            )
        except ProgrammingError:
            # Database not initialized ?
            pass
    return errors


@register()
def check_opp_id_defined(app_configs, **kwargs):
    errors = []
    tropp_observatory_id = settings.TROPP_OBSERVATORY_ID
    if not tropp_observatory_id:
        errors.append(
            Warning(
                "To correctly use OPP you should set TROPP_OBSERVATORY_ID and restart your instance as soon as possible.",
                hint="""
                    TROPP_OBSERVATORY_ID is the identifier of your observatory at national level
                    Ex: TROPP_OBSERVATORY_ID = 20
                """,
                obj=None,
                id="terra_opp.E004",
            )
        )
    return errors


@register()
def check_installed_apps(app_configs, **kwargs):
    errors = []
    if "versatileimagefield" not in settings.INSTALLED_APPS:
        errors.append(
            Error(
                "'terra-opp' needs 'versatileimagefield' in INSTALLED_APPS",
                obj=None,
                id="terra_opp.E001",
            )
        )
    return errors



# Generated by Django 3.2 on 2021-04-13 12:32

from django.db import migrations


def populate_new_identifier(apps, schema_editor):
    Picture = apps.get_model("terra_opp", "Picture")

    for picture in Picture.objects.all():
        if picture.old_identifier:
            picture.identifier = str(picture.old_identifier)
            picture.save()


def populate_old_identifier(apps, schema_editor):
    Picture = apps.get_model("terra_opp", "Picture")

    for picture in Picture.objects.all():
        if picture.identifier:
            picture.old_identifier = int(picture.identifier)
            picture.save()


class Migration(migrations.Migration):

    dependencies = [
        ("terra_opp", "0012_picture_identifier_to_charfield"),
    ]

    operations = [
        migrations.RunPython(
            code=populate_new_identifier,
            reverse_code=populate_old_identifier,
        ),
        migrations.RemoveField(
            model_name="picture",
            name="old_identifier",
        ),
    ]



# Generated by Django 2.2.5 on 2019-10-09 15:07

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("terra_opp", "0001_initial"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="picture",
            name="remarks",
        ),
    ]



# Generated by Django 2.2.5 on 2019-10-09 09:24

from django.conf import settings

try:
    from django.db.models import JSONField
except ImportError:  # TODO Remove when dropping Django releases < 3.1
    from django.contrib.postgres.fields import JSONField
from django.db import migrations, models
import django.db.models.deletion
import versatileimagefield.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("geostore", "0029_auto_20190926_0803"),
    ]

    operations = [
        migrations.CreateModel(
            name="Viewpoint",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("label", models.CharField(max_length=100, verbose_name="Label")),
                (
                    "properties",
                    JSONField(blank=True, default=dict, verbose_name="Properties"),
                ),
                (
                    "point",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="points",
                        to="geostore.Feature",
                    ),
                ),
            ],
            options={
                "ordering": ["-created_at"],
                "permissions": (
                    ("can_download_pdf", "Is able to download a pdf document"),
                ),
            },
        ),
        migrations.CreateModel(
            name="Picture",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("state", models.IntegerField(default=100, verbose_name="State")),
                (
                    "properties",
                    JSONField(blank=True, default=dict, verbose_name="Properties"),
                ),
                (
                    "file",
                    versatileimagefield.fields.VersatileImageField(
                        upload_to="", verbose_name="File"
                    ),
                ),
                ("date", models.DateTimeField(verbose_name="Date")),
                ("remarks", models.TextField(max_length=350, verbose_name="Remarks")),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="pictures",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Owner",
                    ),
                ),
                (
                    "viewpoint",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="pictures",
                        to="terra_opp.Viewpoint",
                    ),
                ),
            ],
            options={
                "ordering": ["-date"],
                "permissions": (
                    ("change_state_picture", "Is able to change the picture state"),
                ),
                "get_latest_by": "date",
            },
        ),
        migrations.CreateModel(
            name="Campaign",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("label", models.CharField(max_length=100, verbose_name="Label")),
                (
                    "assignee",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="assigned_campaigns",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Assigned to",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="campaigns",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="Owner",
                    ),
                ),
                (
                    "viewpoints",
                    models.ManyToManyField(
                        related_name="campaigns", to="terra_opp.Viewpoint"
                    ),
                ),
            ],
            options={
                "ordering": ["-created_at"],
                "permissions": (("manage_all_campaigns", "Can manage all campaigns"),),
            },
        ),
        migrations.AddIndex(
            model_name="picture",
            index=models.Index(
                fields=["viewpoint", "date"], name="terra_opp_p_viewpoi_bbead9_idx"
            ),
        ),
    ]



# Generated by Django 2.2.18 on 2021-02-24 13:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("terra_opp", "0005_auto_20210224_1243"),
    ]

    operations = [
        migrations.AlterField(
            model_name="campaign",
            name="start_date",
            field=models.DateField(verbose_name="Start date"),
        ),
    ]



import csv
import logging
from copy import deepcopy

logger = logging.getLogger(__name__)


class CustomCsvBuilder:
    """Build csv file from dict/list data"""

    def __init__(self, data):
        self.data = data

    def build_header(self, data, parent_key=None):
        header = []

        if isinstance(data, dict):
            for colname, colvalue in data.items():
                full_colname = colname
                if parent_key:
                    full_colname = (
                        parent_key
                        if colname in parent_key
                        else "_".join((parent_key, colname))
                    )

                partial_header = self.build_header(
                    colvalue,
                    parent_key=full_colname,
                )

                header += partial_header

        elif isinstance(data, list):
            if parent_key:
                header.append(parent_key)
            for val in data:
                header += self.build_header(val, parent_key=parent_key)

        else:
            if parent_key:
                header.append(parent_key)
        return set(header)

    def flatten_datatree(self, tree, parent="", sep="_"):
        flatten_tree = []
        if isinstance(tree, dict):
            self._tree_flattend(flatten_tree, tree, parent, sep)
        elif isinstance(tree, list):
            self._list_flattend(flatten_tree, tree, parent, sep)
        else:
            flatten_tree.append({parent: tree})
        return flatten_tree

    def _list_flattend(self, flatten_tree, tree, parent="", sep="_"):
        for i, sibling in enumerate(tree):
            flatten_sibling = self.flatten_datatree(sibling, parent=parent)
            flatten_tree.extend(flatten_sibling)

    def _tree_flattend(self, flatten_tree, tree, parent="", sep="_"):
        for node, child in tree.items():
            parent_node = node if not parent else sep.join((parent, node))
            flatten_child = self.flatten_datatree(child, parent=parent_node)

            if not flatten_tree:
                flatten_tree.extend(flatten_child)
            else:
                for i, sub_child in enumerate(flatten_child):
                    for sub_tree in flatten_tree:
                        if sub_child.keys() <= sub_tree.keys():
                            subtree_copy = deepcopy(sub_tree)
                            subtree_copy.update(sub_child)
                            flatten_tree.append(subtree_copy)
                            break
                    else:
                        try:
                            flatten_tree[i].update(sub_child)
                        except IndexError:
                            flatten_tree.append(sub_child)

    def create_csv(self, output):
        self.csv = output

        header = self.build_header(self.data)
        body = self.flatten_datatree(self.data)

        writer = csv.DictWriter(self.csv, fieldnames=header)

        writer.writeheader()
        for element in body:
            writer.writerow(element)
        output.seek(0)


class Choices:
    """
    Class under GPL Licence from:
    https://github.com/liberation/django-extended-choices/
    Helper class for choices fields in Django.
    A choice value has three representation (constant name, value and
    string). So Choices takes list of such tuples.
    Here is an example of Choices use:
    >>> CHOICES_ALIGNEMENT = Choices(
    ...     ('BAD', 10, u'bad'),
    ...     ('NEUTRAL', 20, u'neutral'),
    ...     ('CHAOTIC_GOOD', 30, u'chaotic good'),
    ...     ('GOOD', 40, u'good'),
    ... )
    >>> CHOICES_ALIGNEMENT.BAD
    10
    >>> CHOICES_ALIGNEMENT.CHOICES_DICT[30]
    u'chaotic good'
    >>> CHOICES_ALIGNEMENT.REVERTED_CHOICES_DICT[u'good']
    40
    As you can see in the above example usage, Choices objects gets three
    attributes:
    - one attribute built after constant names provided in the tuple (like BAD,
      NEUTRAL etc...)
    - a CHOICES_DICT that match value to string
    - a REVERTED_CHOICES_DICT that match string to value
    If you want to create subset of choices, you can
    use the add_subset method
    This method take a name, and then the constants you want to
    have in this subset:
    >>> CHOICES_ALIGNEMENT.add_subset('WESTERN',('BAD', 'GOOD'))
    >>> CHOICES_ALIGNEMENT.WESTERN
    ((10, u'bad'), (40, u'good'))
    """

    def __init__(self, *raw_choices, **kwargs):
        self._CHOICES = None
        self._CHOICES_DICT = None
        self._REVERTED_CHOICES_DICT = None
        self._VALUE_TO_CONST = None
        self._RAW_CHOICES = None
        self._CONSTS = []
        self._CONST_CHOICES = None
        self.parent = None
        if "parent" in kwargs:  # 'subchoice' mode
            self.parent = kwargs["parent"]
            self._CONSTS = kwargs["CONSTS"]
        else:
            name = kwargs.get("name", "CHOICES")  # retrocompatibility
            if name != "CHOICES":  # retrocompatibility
                self._RAW_CHOICES = tuple()
                self.add_choices(name, *raw_choices)
            else:
                self._RAW_CHOICES = raw_choices  # default behaviour
                self._make_consts(*raw_choices)

    def _make_consts(self, *raw_choices):
        for choice in raw_choices:
            const, value, string = choice
            if hasattr(self, const):
                raise ValueError(
                    "You cannot declare two constants "
                    "with the same name! %s " % str(choice)
                )
            if value in [getattr(self, c) for c in self._CONSTS]:
                raise ValueError(
                    "You cannot declare two constants "
                    "with the same value! %s " % str(choice)
                )
            setattr(self, const, value)
            self._CONSTS.append(const)

            # retrocompatibilty
            self._CHOICES = None
            self._CHOICES_DICT = None
            self._REVERTED_CHOICES_DICT = None

    def add_choices(self, name="CHOICES", *raw_choices):
        # for retrocompatibility
        # we may have to call _build_choices
        # more than one time and so append the
        # new choices to the already existing ones
        RAW_CHOICES = list(self._RAW_CHOICES)
        self._RAW_CHOICES = tuple(RAW_CHOICES + list(raw_choices))
        self._make_consts(*raw_choices)
        if name != "CHOICES":
            # for retrocompatibility
            # we make a subset with new choices
            constants_for_subset = []
            for choice in raw_choices:
                const, value, string = choice
                constants_for_subset.append(const)
            self.add_subset(name, constants_for_subset)

    def add_subset(self, name, constants):
        if hasattr(self, name):
            raise ValueError(
                "Cannot use %s as a subset name." "It's already an attribute." % name
            )

        subset = Choices(parent=self, CONSTS=constants)
        setattr(self, name, subset)

        # For retrocompatibility
        setattr(self, "%s_DICT" % name, getattr(subset, "CHOICES_DICT"))
        setattr(
            self, "REVERTED_%s_DICT" % name, getattr(subset, "REVERTED_CHOICES_DICT")
        )

    @property
    def RAW_CHOICES(self):
        if self._RAW_CHOICES is None:
            if self.parent:
                self._RAW_CHOICES = tuple(
                    (c, k, v)
                    for c, k, v in self.parent.RAW_CHOICES
                    if c in self._CONSTS
                )
            else:
                raise ValueError(
                    "Implementation problem : first "
                    "ancestor should have a _RAW_CHOICES"
                )
        return self._RAW_CHOICES

    @property
    def CHOICES(self):
        """
        Tuple of tuples (value, display_value).
        """
        if self._CHOICES is None:
            self._CHOICES = tuple(
                (k, v) for c, k, v in self.RAW_CHOICES if c in self._CONSTS
            )
        return self._CHOICES

    @property
    def CHOICES_DICT(self):
        if self._CHOICES_DICT is None:
            self._CHOICES_DICT = {}
            for c, k, v in self.RAW_CHOICES:
                if c in self._CONSTS:
                    self._CHOICES_DICT[k] = v
        return self._CHOICES_DICT

    @property
    def REVERTED_CHOICES_DICT(self):
        """
        Dict {"display_value": "value"}
        """
        # FIXME: rename in a more friendly name, like STRING_TO_VALUE?
        if self._REVERTED_CHOICES_DICT is None:
            self._REVERTED_CHOICES_DICT = {}
            for c, k, v in self.RAW_CHOICES:
                if c in self._CONSTS:
                    self._REVERTED_CHOICES_DICT[v] = k
        return self._REVERTED_CHOICES_DICT

    @property
    def VALUE_TO_CONST(self):
        """
        Dict {"value": "const"}
        """
        if self._VALUE_TO_CONST is None:
            self._VALUE_TO_CONST = {}
            for c, k, v in self.RAW_CHOICES:
                if c in self._CONSTS:
                    self._VALUE_TO_CONST[k] = c
        return self._VALUE_TO_CONST

    @property
    def CONST_CHOICES(self):
        """
        Tuple of tuples (constant, display_value).
        """
        if self._CONST_CHOICES is None:
            self._CONST_CHOICES = tuple(
                (c, v) for c, k, v in self.RAW_CHOICES if c in self._CONSTS
            )
        return self._CONST_CHOICES

    def __contains__(self, item):
        """
        Make smarter to check if a value is valid for a Choices.
        """
        return item in self.CHOICES_DICT

    def __iter__(self):
        return self.CHOICES.__iter__()

    def __eq__(self, other):
        return other == self.CHOICES

    def __ne__(self, other):
        return other != self.CHOICES

    def __len__(self):
        return self.CHOICES.__len__()

    def __add__(self, item):
        """
        Needed to make MY_CHOICES + OTHER_CHOICE
        """
        if not isinstance(item, (Choices, tuple)):
            raise ValueError(
                "This operand could only by evaluated "
                "with Choices or tuple instances. "
                "Got %s instead." % type(item)
            )
        return self.CHOICES + tuple(item)

    def __repr__(self):
        return self.CHOICES.__repr__()



from .models import Picture, Campaign
from rest_framework.permissions import SAFE_METHODS, BasePermission


class ViewpointPermission(BasePermission):
    """Read only for anonymous users or has terra permission"""

    def has_permission(self, request, view):
        if request.method in SAFE_METHODS:
            return True
        return not request.user.is_anonymous and request.user.has_terra_perm(
            "can_manage_viewpoints"
        )


class CampaignPermission(BasePermission):
    """Read only for authenticated users or has terra permission"""

    def has_permission(self, request, view):

        # Allow all sheets download
        if "all_sheets" in request.path:
            return True

        if request.user.is_anonymous:
            return False

        if request.method in SAFE_METHODS:
            return request.user.has_terra_perm(
                "can_add_pictures"
            ) or request.user.has_terra_perm("can_manage_campaigns")

        return request.user.has_terra_perm("can_manage_campaigns")

    def has_object_permission(self, request, view, obj):
        """
        User with permission can_add_pictures can only
        modify assigned campaings
        """

        # Allow all sheets download
        if "all_sheets" in request.path:
            return True

        if request.method in SAFE_METHODS:
            if request.user.has_terra_perm("can_manage_campaigns"):
                return True
            elif request.user.has_terra_perm("can_add_pictures"):
                return obj.assignee == request.user and obj.state != Campaign.DRAFT

        return request.user.has_terra_perm("can_manage_campaigns")


class PicturePermission(BasePermission):
    """Read only for anonymous users or has terra permissions"""

    def has_permission(self, request, view):

        if request.method in SAFE_METHODS:
            return True

        if request.user.is_anonymous:
            return False

        if request.user.has_terra_perm("can_manage_pictures"):
            return True

        if request.user.has_terra_perm("can_add_pictures"):
            return True

        return False

    def has_object_permission(self, request, view, obj):
        """
        User with permission can_add_pictures can only
        modify self picture in state DRAFT or REFUSED
        """
        if request.method in SAFE_METHODS:
            return True

        if request.user.has_terra_perm("can_manage_pictures"):
            return True

        if request.user.has_terra_perm("can_add_pictures"):
            return obj.owner == request.user and obj.state in [
                Picture.DRAFT,
                Picture.REFUSED,
            ]

        return False



from django.conf import settings
from django.core.management import BaseCommand
from geostore import GeometryTypes
from geostore.models import Layer


class Command(BaseCommand):
    help = "Create a django-geostore point layer to store observatory viewpoints"

    def add_arguments(self, parser):
        parser.add_argument(
            "-n", "--name", action="store", help="Name for layer to create."
        )
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="Create even if the TROPP_OBSERVATORY_LAYER_PK already set.",
        )

    def handle(self, *args, **options):
        existing_observatory = getattr(settings, "TROPP_OBSERVATORY_LAYER_PK")

        name = options.get("name") or "opp_baselayer"
        force = options.get("force")

        if existing_observatory and not force:
            try:
                Layer.objects.get(pk=existing_observatory)
            except Layer.DoesNotExist:
                self.stdout.write(
                    self.style.WARNING(
                        f"Defined layer with PK {existing_observatory} defined in setting TROPP_OBSERVATORY_LAYER_PK does not exists in database (-f to force createn)"
                    )
                )
            else:
                self.stdout.write(
                    self.style.SUCCESS(
                        f"A layer already exists for this TROPP_OBSERVATORY_LAYER_PK: {existing_observatory} - Do nothing"
                    )
                )
        else:
            if force:
                try:
                    # whether a layer with this id already exists
                    Layer.objects.get(pk=existing_observatory)
                    self.stdout.write(
                        self.style.WARNING(
                            f"A layer already exists for this TROPP_OBSERVATORY_LAYER_PK: {existing_observatory} - Do nothing"
                        )
                    )
                    return
                except Layer.DoesNotExist:
                    pass

            try:
                # whether a layer with this name already exists
                layer = Layer.objects.get(name=name)
                self.stdout.write(
                    self.style.WARNING(
                        f"A layer with pk {layer} already exists for this name: {name} - Can't create new one."
                    )
                )
                return
            except Layer.DoesNotExist:
                pass

            layer = Layer.objects.create(geom_type=GeometryTypes.Point, name=f"{name}")
            self.stdout.write(
                self.style.SUCCESS(
                    f"Layer has been created. Please set TROPP_OBSERVATORY_LAYER_PK={layer.pk} in project settings"
                )
            )



# Generated by Django 3.1.7 on 2021-04-12 10:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("terra_opp", "0009_picture_campaign"),
    ]

    operations = [
        migrations.AddField(
            model_name="picture",
            name="identifier",
            field=models.PositiveIntegerField(null=True, verbose_name="Identifier"),
        ),
    ]



# Generated by Django 2.2.18 on 2021-04-27 12:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("terra_opp", "0013_rename_picture_newidentifier"),
    ]

    operations = [
        migrations.AddField(
            model_name="theme",
            name="category",
            field=models.CharField(default="", max_length=100, verbose_name="Category"),
        ),
    ]



# Generated by Django 3.1.7 on 2021-03-10 16:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("terra_opp", "0003_auto_20200630_1315"),
    ]

    operations = [
        migrations.AddField(
            model_name="viewpoint",
            name="active",
            field=models.BooleanField(default=True, verbose_name="Active"),
        ),
    ]



# Generated by Django 2.2.18 on 2021-03-25 10:22

from django.db import migrations, models
import terra_opp.models
import versatileimagefield.fields


class Migration(migrations.Migration):

    dependencies = [
        ("terra_opp", "0006_auto_20210224_1333"),
    ]

    operations = [
        migrations.RenameField(
            model_name="picture",
            old_name="state",
            new_name="old_state",
        ),
        migrations.AlterModelOptions(
            name="campaign",
            options={"ordering": ["-start_date", "-created_at"]},
        ),
        migrations.AddField(
            model_name="campaign",
            name="state",
            field=models.CharField(
                choices=[
                    ("draft", "Draft"),
                    ("started", "Started"),
                    ("closed", "Closed"),
                ],
                default="draft",
                max_length=10,
                verbose_name="State",
            ),
        ),
        migrations.AddField(
            model_name="picture",
            name="state",
            field=models.CharField(
                choices=[
                    ("draft", "Draft"),
                    ("submited", "Submited"),
                    ("accepted", "Accepted"),
                    ("refused", "Refused"),
                ],
                default="draft",
                max_length=10,
                verbose_name="State",
            ),
        ),
        migrations.AlterField(
            model_name="campaign",
            name="label",
            field=models.CharField(max_length=100, verbose_name="Label"),
        ),
        migrations.AlterField(
            model_name="picture",
            name="file",
            field=versatileimagefield.fields.VersatileImageField(
                upload_to=terra_opp.models.image_upload_to, verbose_name="File"
            ),
        ),
    ]



from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.gis.db import models
from pathlib import Path

try:
    from django.db.models import JSONField
except ImportError:  # TODO Remove when dropping Django releases < 3.1
    from django.contrib.postgres.fields import JSONField

from django.utils.translation import ugettext_lazy as _
from versatileimagefield.fields import VersatileImageField

from terra_settings.mixins import BaseUpdatableModel
from geostore.models import Feature
from .signals import state_change


# from django.db.models import Count


class BaseLabelModel(BaseUpdatableModel):
    label = models.CharField(_("Label"), max_length=100)

    class Meta:
        abstract = True

    def __str__(self):
        return f"{self.label}"


class ViewpointsManager(models.Manager):
    def with_accepted_pictures(self):
        return (
            super()
            .get_queryset()
            .filter(
                pictures__state=Picture.ACCEPTED,
            )
            .distinct()
        )


class City(BaseLabelModel):
    class Meta:
        verbose_name_plural = _("Cities")


class Theme(BaseLabelModel):
    category = models.CharField(_("Category"), max_length=100, default="")


class Viewpoint(BaseLabelModel):
    point = models.ForeignKey(
        Feature,
        on_delete=models.CASCADE,
        related_name="points",
    )
    city = models.ForeignKey(
        City,
        on_delete=models.CASCADE,
        related_name="viewpoints",
        null=True,
    )
    themes = models.ManyToManyField(
        Theme,
        related_name="viewpoints",
    )
    properties = JSONField(_("Properties"), default=dict, blank=True)
    related = GenericRelation("datastore.RelatedDocument")
    active = models.BooleanField(_("Active"), default=True)

    objects = ViewpointsManager()

    @property
    def ordered_pics(self):
        if hasattr(self, "_ordered_pics"):
            # if _ordered_pics set by prefetch on qs, get it
            return self._ordered_pics
        else:
            # not prefetch, compute
            return self.pictures.order_by("-created_at")

    @property
    def ordered_pics_by_date(self):
        return self.pictures.order_by("date")

    @property
    def picture(self):
        pics = self.ordered_pics
        return pics[0].file if pics else None

    class Meta:
        permissions = (("can_download_pdf", "Is able to download a pdf document"),)
        ordering = ["-created_at"]


class CampaignManager(models.Manager):
    # Add stats to campaign
    def with_stats(self):
        return self.annotate(
            viewpoints_total=models.Count("viewpoints", distinct=True),
            pictures_submited=models.Count(
                "pictures__pk",
                filter=models.Q(pictures__state=Picture.SUBMITTED),
                distinct=True,
            ),
            pictures_accepted=models.Count(
                "pictures__pk",
                filter=models.Q(pictures__state=Picture.ACCEPTED),
                distinct=True,
            ),
            pictures_missing=models.F("viewpoints_total")
            - models.F("pictures_submited")
            - models.F("pictures_accepted"),
        )


class Campaign(BaseLabelModel):
    objects = CampaignManager()

    DRAFT = "draft"
    STARTED = "started"
    CLOSED = "closed"
    STATES = (
        (DRAFT, _("Draft")),
        (STARTED, _("Started")),
        (CLOSED, _("Closed")),
    )

    start_date = models.DateField(_("Start date"))
    viewpoints = models.ManyToManyField(
        Viewpoint,
        related_name="campaigns",
    )
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        verbose_name=_("Owner"),
        related_name="campaigns",
    )
    assignee = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        verbose_name=_("Assigned to"),
        related_name="assigned_campaigns",
    )
    state = models.CharField(_("State"), default=DRAFT, max_length=10, choices=STATES)

    # Auto close campaign if all pictures are accepted
    def check_state(self):
        if (
            self.pictures.filter(state=Picture.ACCEPTED).count()
            == self.viewpoints.count()
        ):
            self.state = Campaign.CLOSED
            self.save()

    class Meta:
        ordering = ["-start_date", "-created_at"]

    def save(self, *args, **kwargs):
        prev_state = None
        if self.id:
            prev_state = Campaign.objects.get(id=self.id).state

        super().save(*args, **kwargs)

        # Send state changed event
        if prev_state != self.state:
            state_change.send(
                sender=Campaign,
                instance=self,
                prev_state=prev_state,
                new_state=self.state,
            )


def image_upload_to(instance, filename):
    date_str = instance.date.strftime("%Y-%m-%d_%H-%M-%S")
    filename = Path(filename)
    return f"pictures/viewpoint_{instance.viewpoint.id}/{date_str}{filename.suffix}"


class Picture(BaseUpdatableModel):

    DRAFT = "draft"
    SUBMITED = "submited"  # Typo...
    SUBMITTED = "submited"
    ACCEPTED = "accepted"
    REFUSED = "refused"

    STATES = (
        (DRAFT, _("Draft")),
        (SUBMITTED, _("Submitted")),
        (ACCEPTED, _("Accepted")),
        (REFUSED, _("Refused")),
    )

    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        verbose_name=_("Owner"),
        related_name="pictures",
    )
    viewpoint = models.ForeignKey(
        Viewpoint,
        on_delete=models.CASCADE,
        related_name="pictures",
    )
    campaign = models.ForeignKey(
        Campaign,
        null=True,
        on_delete=models.SET_NULL,
        related_name="pictures",
        default=None,
    )
    # States may be : draft, submitted, accepted, refused
    state = models.CharField(_("State"), default=DRAFT, choices=STATES, max_length=10)

    properties = JSONField(_("Properties"), default=dict, blank=True)
    file = VersatileImageField(_("File"), upload_to=image_upload_to)

    # Different from created_at which is the upload date
    date = models.DateTimeField(_("Date"))
    identifier = models.CharField(_("Identifier"), default="", max_length=10)

    class Meta:
        permissions = (
            ("change_state_picture", "Is able to change the picture " "state"),
        )
        # It's our main way of sorting pictures, so it better be indexed
        indexes = [
            models.Index(fields=["viewpoint", "date"]),
        ]
        get_latest_by = "date"
        ordering = ["-date"]

    def save(self, *args, **kwargs):
        prev_state = None
        if self.id:
            prev_state = Picture.objects.get(id=self.id).state

        super().save(*args, **kwargs)
        # Update Campaign status if any
        if self.campaign:
            self.campaign.check_state()

        # Send state changed event
        if prev_state != self.state:
            state_change.send(
                sender=Picture,
                instance=self,
                prev_state=prev_state,
                new_state=self.state,
            )

    def get_identifier(self):
        obs_id = settings.TROPP_OBSERVATORY_ID or ""
        pic_index = list(
            self.viewpoint.ordered_pics_by_date.values_list("id", flat=True)
        ).index(self.id)
        pic_index += 1  # list index start at 0, picture order start at 1
        return f"{obs_id}0{self.viewpoint.id:03}{pic_index:02}"



from django.shortcuts import resolve_url
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APITestCase
from django.utils import timezone
from django.core.files.uploadedfile import SimpleUploadedFile

from terra_accounts.tests.factories import TerraUserFactory
from terra_opp.tests.factories import CampaignFactory, ViewpointFactory, PictureFactory
from terra_opp.tests.mixins import TestPermissionsMixin


class CampaignTestCase(TestPermissionsMixin, APITestCase):
    def setUp(self):
        self.photograph = TerraUserFactory()
        self.user = TerraUserFactory()

    def _gen_file(self):
        return SimpleUploadedFile(
            name="add_picture_test.jpg",
            content=open(
                "terra_opp/tests/placeholder.jpg",
                "rb",
            ).read(),
            content_type="image/jpeg",
        )

    def _authent(self, user, permissions):
        self.client.force_authenticate(user=user)
        self._set_permissions(
            permissions,
            user,
        )

    def as_photograph(self):
        self._authent(self.photograph, ["can_add_pictures"])

    def as_admin(self):
        self._authent(self.user, ["can_manage_campaigns", "can_manage_pictures"])

    def test_list_campaign(self):
        viewpoint = ViewpointFactory()
        viewpoint2 = ViewpointFactory()
        campaign = CampaignFactory(assignee=self.photograph, state="started")
        campaign.viewpoints.set([viewpoint, viewpoint2])

        campaign_other = CampaignFactory(
            state="started"
        )  # campaign for other photograph

        campaign_same_not_started = CampaignFactory(
            state="draft", assignee=self.photograph
        )  # campaign for same photograph but not started

        list_url = reverse("terra_opp:campaign-list")
        campaign_url = resolve_url("terra_opp:campaign-detail", pk=campaign.pk)
        campaign_other_url = resolve_url(
            "terra_opp:campaign-detail", pk=campaign_other.pk
        )
        campaign_not_started_url = resolve_url(
            "terra_opp:campaign-detail", pk=campaign_same_not_started.pk
        )

        # First we try as anonymous
        self.assertIn(
            self.client.get(list_url).status_code,
            [status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN],
        )
        self.assertIn(
            self.client.get(campaign_url).status_code,
            [status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN],
        )
        self.assertIn(
            self.client.get(campaign_other_url).status_code,
            [status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN],
        )

        # Then we try with can_add_pictures permission
        self.as_photograph()

        response = self.client.get(list_url)

        self.assertEqual(status.HTTP_200_OK, response.status_code)
        # Only one should be visible
        self.assertEqual(1, response.data.get("count"))

        response = self.client.get(campaign_url)
        self.assertEqual(status.HTTP_200_OK, response.status_code)

        # Others campaigns should be forbidden
        self.assertEqual(
            status.HTTP_403_FORBIDDEN, self.client.get(campaign_other_url).status_code
        )
        # Not started campaigns should forbidden
        self.assertEqual(
            status.HTTP_403_FORBIDDEN,
            self.client.get(campaign_not_started_url).status_code,
        )

        # Then as admin
        self.as_admin()

        response = self.client.get(list_url)
        self.assertEqual(status.HTTP_200_OK, response.status_code)
        self.assertEqual(response.data.get("count"), 3)

        campaign_response = self.client.get(campaign_url)
        self.assertEqual(status.HTTP_200_OK, campaign_response.status_code)

        self.assertNumQueries(3, lambda: self.client.get(campaign_url))

        self.assertEqual(
            status.HTTP_200_OK, self.client.get(campaign_other_url).status_code
        )
        self.assertEqual(
            status.HTTP_200_OK, self.client.get(campaign_not_started_url).status_code
        )

    def test_search_campaign(self):
        viewpoint = ViewpointFactory(pictures__state="accepted")

        campaign = CampaignFactory(assignee=self.photograph)
        campaign.viewpoints.set([viewpoint])

        CampaignFactory(assignee=self.photograph, state="started")
        CampaignFactory(assignee=self.photograph, state="closed")

        picture = PictureFactory(state="draft", viewpoint=viewpoint, campaign=campaign)

        list_url = resolve_url("terra_opp:campaign-list")

        self.as_admin()

        response = self.client.get(list_url, {"state": "draft"})
        self.assertEqual(1, response.data.get("count"))

        response = self.client.get(list_url, {"state": "started"})
        self.assertEqual(1, response.data.get("count"))

        response = self.client.get(list_url, {"state": "closed"})
        self.assertEqual(1, response.data.get("count"))

        response = self.client.get(list_url, {"state": "foo"})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        response = self.client.get(list_url, {"pictures__state": "draft"})
        self.assertEqual(1, response.data.get("count"))
        response = self.client.get(list_url, {"pictures__state": "accepted"})
        self.assertEqual(0, response.data.get("count"))

        picture.state = "accepted"
        picture.save()

        response = self.client.get(list_url, {"pictures__state": "draft"})
        self.assertEqual(0, response.data.get("count"))
        response = self.client.get(list_url, {"pictures__state": "accepted"})
        self.assertEqual(1, response.data.get("count"))

        response = self.client.get(list_url, {"pictures__state": "foo"})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_post_campaign(self):
        data = {
            "label": "My campaign",
            "start_date": "2021-01-25",
            "assignee": self.photograph.pk,
            "viewpoints": [ViewpointFactory().pk, ViewpointFactory().pk],
        }

        self.as_photograph()

        response = self.client.post(reverse("terra_opp:campaign-list"), data)
        self.assertEqual(status.HTTP_403_FORBIDDEN, response.status_code)

        self.as_admin()

        response = self.client.post(reverse("terra_opp:campaign-list"), data)

        self.assertEqual(status.HTTP_201_CREATED, response.status_code)

    def test_add_picture(self):
        viewpoint = ViewpointFactory()
        viewpoint2 = ViewpointFactory()
        viewpoint3 = ViewpointFactory()
        viewpoint4 = ViewpointFactory()
        viewpoint5 = ViewpointFactory()
        viewpoint6 = ViewpointFactory()

        campaign1 = CampaignFactory(assignee=self.photograph, state="started")
        campaign1.viewpoints.set([viewpoint, viewpoint2])
        initial_picture_count1 = viewpoint.pictures.count()

        campaign2 = CampaignFactory(assignee=self.photograph, state="draft")
        campaign2.viewpoints.set([viewpoint3, viewpoint4])

        campaign3 = CampaignFactory(state="started")
        campaign3.viewpoints.set([viewpoint5, viewpoint6])

        data = {
            "viewpoint": viewpoint.pk,
            "date": timezone.datetime(2020, 8, 19, tzinfo=timezone.utc),
            "file": self._gen_file(),
            "state": "draft",
        }

        self.as_photograph()

        # Nominal cases should works
        response = self.client.post(
            reverse("terra_opp:picture-list"),
            data,
            format="multipart",
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(viewpoint.pictures.count(), initial_picture_count1 + 1)
        self.assertEqual(
            viewpoint.pictures.latest().date,
            timezone.datetime(2020, 8, 19, tzinfo=timezone.utc),
        )
        self.assertEqual(
            viewpoint.pictures.latest().state,
            "draft",
        )

        # Should not allow to create second photo
        data["file"] = self._gen_file()
        response = self.client.post(
            reverse("terra_opp:picture-list"),
            data,
            format="multipart",
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        # Should update picture but ignore accepted state
        response = self.client.patch(
            reverse(
                "terra_opp:picture-detail",
                args=[viewpoint.pictures.latest().pk],
            ),
            {"state": "accepted"},
            format="multipart",
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data.get("state"), "draft")

        # Should update previous picture and change state
        response = self.client.patch(
            reverse(
                "terra_opp:picture-detail",
                args=[viewpoint.pictures.latest().pk],
            ),
            {"state": "submited"},
            format="multipart",
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # Should not allow modification on submitted picture
        response = self.client.patch(
            reverse(
                "terra_opp:picture-detail",
                args=[viewpoint.pictures.latest().pk],
            ),
            {"state": "accepted"},
            format="multipart",
        )

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        # Post to draft campaign
        data["viewpoint"] = viewpoint3.pk
        data["file"] = self._gen_file()
        response = self.client.post(
            reverse("terra_opp:picture-list"),
            data,
            format="multipart",
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        # Post to started campaign for another assignee
        data["viewpoint"] = viewpoint5.pk
        data["file"] = self._gen_file()
        response = self.client.post(
            reverse("terra_opp:picture-list"),
            data,
            format="multipart",
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        # Post to started campaign but with bad viewpoint
        data["viewpoint"] = viewpoint5.pk
        data["file"] = self._gen_file()
        data["campaign"] = campaign1.pk
        response = self.client.post(
            reverse("terra_opp:picture-list"),
            data,
            format="multipart",
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        # Post to started campaign bad picture state
        data["viewpoint"] = viewpoint5.pk
        data["file"] = self._gen_file()
        data["campaign"] = campaign1.pk
        response = self.client.post(
            reverse("terra_opp:picture-list"),
            data,
            format="multipart",
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_workflow(self):
        viewpoint = ViewpointFactory()
        viewpoint2 = ViewpointFactory()

        campaign1 = CampaignFactory(assignee=self.photograph, state="started")
        campaign1.viewpoints.set([viewpoint, viewpoint2])

        data = {
            "viewpoint": viewpoint.pk,
            "date": timezone.datetime(2020, 8, 19, tzinfo=timezone.utc),
            "file": self._gen_file(),
            "state": "draft",
        }

        # Photograph post draft photo
        self.as_photograph()
        self.client.post(
            reverse("terra_opp:picture-list"),
            data,
            format="multipart",
        )

        latest_picture = viewpoint.pictures.latest()

        # Then submit the photo
        self.client.patch(
            reverse(
                "terra_opp:picture-detail",
                args=[latest_picture.pk],
            ),
            {"state": "submited"},
        )

        # Admin refuse the photo
        self.as_admin()
        self.client.patch(
            reverse(
                "terra_opp:picture-detail",
                args=[latest_picture.pk],
            ),
            {"state": "refused"},
        )

        # Photograph update the photo
        self.as_photograph()
        self.client.patch(
            reverse(
                "terra_opp:picture-detail",
                args=[latest_picture.pk],
            ),
            {"state": "submited", "file": self._gen_file()},
            format="multipart",
        )

        # Admin accept the photo
        self.as_admin()
        self.client.patch(
            reverse(
                "terra_opp:picture-detail",
                args=[latest_picture.pk],
            ),
            {"state": "accepted"},
        )

        latest_picture.refresh_from_db()
        self.assertEqual(latest_picture.state, "accepted")

    def test_statistics(self):
        viewpoint = ViewpointFactory(pictures__state="accepted")
        viewpoint2 = ViewpointFactory(pictures__state="accepted")
        viewpoint3 = ViewpointFactory(pictures__state="accepted")
        viewpoint4 = ViewpointFactory(pictures__state="accepted")

        campaign = CampaignFactory(assignee=self.photograph, state="started")
        campaign.viewpoints.set([viewpoint, viewpoint2, viewpoint3, viewpoint4])
        campaign_url = resolve_url("terra_opp:campaign-detail", pk=campaign.pk)

        data = {
            "viewpoint": viewpoint.pk,
            "date": timezone.datetime(2020, 8, 19, tzinfo=timezone.utc),
            "file": self._gen_file(),
            "state": "draft",
        }

        # Stats without any photo
        self.as_admin()

        response = self.client.get(campaign_url)

        def stats(response):
            return response.json().get("statistics")

        self.assertEqual(
            stats(response),
            {"total": 4, "missing": 4, "submited": 0, "accepted": 0},
        )

        # Add draft photo
        self.as_photograph()
        response = self.client.post(
            reverse("terra_opp:picture-list"),
            data,
            format="multipart",
        )

        # Draft photo are considered as missing...
        self.as_admin()
        response = self.client.get(campaign_url)
        self.assertEqual(
            stats(response),
            {"total": 4, "missing": 4, "submited": 0, "accepted": 0},
        )

        # ...But are listed
        self.assertEqual(
            response.json()["pictures"],
            [
                {
                    "viewpoint": 14,
                    "date": "2020-08-19T00:00:00Z",
                    "id": 22,
                    "state": "draft",
                }
            ],
        )

        # Submit photo
        self.as_photograph()
        latest_picture = viewpoint.pictures.latest()
        response = self.client.patch(
            reverse(
                "terra_opp:picture-detail",
                args=[latest_picture.pk],
            ),
            {"state": "submited"},
        )

        # Should have 1 submited
        self.as_admin()
        response = self.client.get(campaign_url)
        self.assertEqual(
            stats(response),
            {"total": 4, "missing": 3, "submited": 1, "accepted": 0},
        )

        # Submit another photo
        self.as_photograph()
        data["viewpoint"] = viewpoint2.pk
        data["file"] = self._gen_file()
        data["state"] = "submited"
        response = self.client.post(
            reverse("terra_opp:picture-list"),
            data,
            format="multipart",
        )

        # Should have 2 submited
        self.as_admin()
        response = self.client.get(campaign_url)
        self.assertEqual(
            stats(response),
            {"total": 4, "missing": 2, "submited": 2, "accepted": 0},
        )

        response = self.client.patch(
            reverse(
                "terra_opp:picture-detail",
                args=[latest_picture.pk],
            ),
            {"state": "accepted"},
        )

        # One shoud be accepted
        response = self.client.get(campaign_url)
        self.assertEqual(
            stats(response),
            {"total": 4, "missing": 2, "submited": 1, "accepted": 1},
        )
        self.assertEqual(
            response.json()["pictures"],
            [
                {
                    "viewpoint": 15,
                    "date": "2020-08-19T00:00:00Z",
                    "id": 23,
                    "state": "submited",
                },
                {
                    "viewpoint": 14,
                    "date": "2020-08-19T00:00:00Z",
                    "id": 22,
                    "state": "accepted",
                },
            ],
        )

        # Perf check 3 queries even if 2 images and 4 viewpoints
        # One for campaign, one for viewpoints, one for pictures
        self.assertNumQueries(3, lambda: self.client.get(campaign_url))

        latest_picture2 = viewpoint2.pictures.latest()

        response = self.client.patch(
            reverse(
                "terra_opp:picture-detail",
                args=[latest_picture2.pk],
            ),
            {"state": "refused"},
        )

        # Refused are considered as missing
        response = self.client.get(campaign_url)
        self.assertEqual(
            stats(response),
            {"total": 4, "missing": 3, "submited": 0, "accepted": 1},
        )

    def test_auto_close(self):
        viewpoint = ViewpointFactory(pictures__state="accepted")
        viewpoint2 = ViewpointFactory(pictures__state="accepted")

        campaign = CampaignFactory(assignee=self.photograph, state="started")
        campaign.viewpoints.set([viewpoint, viewpoint2])
        campaign_url = resolve_url("terra_opp:campaign-detail", pk=campaign.pk)

        data = {
            "viewpoint": viewpoint.pk,
            "date": timezone.datetime(2020, 8, 19, tzinfo=timezone.utc),
            "file": self._gen_file(),
            "state": "draft",
        }

        # Stats without any photo
        self.as_admin()

        response = self.client.get(campaign_url)

        def stats(response):
            return response.json().get("statistics")

        # Add draft photo
        self.as_photograph()
        response = self.client.post(
            reverse("terra_opp:picture-list"),
            data,
            format="multipart",
        )
        data["viewpoint"] = viewpoint2.pk
        data["file"] = self._gen_file()
        response = self.client.post(
            reverse("terra_opp:picture-list"),
            data,
            format="multipart",
        )

        self.as_admin()

        latest_picture = viewpoint.pictures.latest()
        latest_picture2 = viewpoint2.pictures.latest()

        response = self.client.patch(
            reverse(
                "terra_opp:picture-detail",
                args=[latest_picture.pk],
            ),
            {"state": "accepted"},
        )

        # Should still be started
        response = self.client.get(campaign_url)
        self.assertEqual(
            response.json()["state"],
            "started",
        )

        response = self.client.patch(
            reverse(
                "terra_opp:picture-detail",
                args=[latest_picture2.pk],
            ),
            {"state": "accepted"},
        )

        # Should be closed after last accepted picture
        response = self.client.get(campaign_url)
        self.assertEqual(
            response.json()["state"],
            "closed",
        )



from django import template
from versatileimagefield.files import VersatileImageFieldFile
from versatileimagefield.utils import get_url_from_image_key

register = template.Library()


@register.filter
def as_versatile(file, image_key):
    """
    :param file: a FileField value
    :param image_key: a versatile image key
    :return: a sized image url for a normal file

    Example: {{ instance.document|as_versatile:'thumbnail__750x1500' }}
    """
    file.field.ppoi_field = None  # hack to make this work
    versatile_file = VersatileImageFieldFile(file, file.field, file.name)
    return get_url_from_image_key(versatile_file, image_key)



from PIL import Image
from django.conf import settings
from django.http import HttpRequest
from versatileimagefield.serializers import VersatileImageFieldSerializer
from terra_opp.models import Picture, Viewpoint


def update_point_properties(viewpoint: Viewpoint, request: HttpRequest):
    """
    Update the point properties of the given Viewpoint instance.
    The properties are the viewpoint's id and label, and all those defined in the
    TROPP_FEATURES_PROPERTIES_FROM_VIEWPOINT setting.

    :param Viewpoint viewpoint: Viewpoint instance used to update its point properties
    :param HttpRequest request: The original request instance.
    """
    point = viewpoint.point
    properties = {
        "viewpoint_id": viewpoint.id,
        "viewpoint_label": viewpoint.label,
        "viewpoint_city": viewpoint.city.label if viewpoint.city else "",
        "viewpoint_active": viewpoint.active,
    }
    # Merging the properties bellow in the ones already present in the point
    point.properties = {**point.properties, **properties}

    # Add thumbnail representation in the feature's properties
    if viewpoint.pictures.filter(state="accepted").exists():
        change_point_thumbnail(
            viewpoint.pictures.filter(state="accepted").latest(),
            context={"request": request},
        )

    # Add any specified viewpoint property in the feature's properties
    for prop in settings.TROPP_FEATURES_PROPERTIES_FROM_VIEWPOINT:
        value = viewpoint.properties.get(prop)
        if value is not None:
            point.properties[f"viewpoint_{prop}"] = value

    point.save()


def change_point_thumbnail(picture: Picture, context):
    """
    Change the point thumbnail of the given Picture instance. The context is mandatory as it will be used to determine
    the picture location.

    :param Picture picture: Picture instance used to update its related point thumbnail property.
    :param context: The context containing the original request instance.
    :return:
    """
    """ Change the picture's point thumbnail """

    terra_opp_versatile_serializer = VersatileImageFieldSerializer("terra_opp")
    terra_opp_versatile_serializer._context = context
    last_picture_sizes = terra_opp_versatile_serializer.to_representation(picture.file)

    # Use best image thumbnail
    with Image.open(picture.file) as pic:
        width, height = pic.size
        if width < height:
            picture.viewpoint.point.properties[
                "viewpoint_picture"
            ] = last_picture_sizes["thumbnail_vertical"]
        else:
            picture.viewpoint.point.properties[
                "viewpoint_picture"
            ] = last_picture_sizes["thumbnail"]

    picture.viewpoint.point.save()


def remove_point_thumbnail(picture: Picture, request: HttpRequest):
    """
    Remove and/or update the point's thumbnail of the given Picture instance only if the picture was the latest.
    The request is mandatory as it will be used to determine the picture location.

    :param Picture picture: Picture instance used to update its related point thumbnail property.
    :param HttpRequest request: The original request instance.
    """
    latest_picture = picture.viewpoint.pictures.latest()

    if latest_picture == picture and picture.viewpoint.pictures.count() > 1:
        # we need to update the point thumbnail
        new_latest_picture = picture.viewpoint.ordered_pics[1]
        change_point_thumbnail(new_latest_picture, context={"request": request})
    elif "viewpoint_picture" in picture.viewpoint.point.properties:
        # just remove the viewpoint_picture from point properties
        picture.viewpoint.point.properties.pop("viewpoint_picture")
        picture.viewpoint.point.save()



from pprint import pformat

import coreapi
import coreschema
from django.conf import settings
from django.contrib.postgres.fields.jsonb import KeyTextTransform
from django_filters import FilterSet
from django_filters.rest_framework import filters
from rest_framework.filters import BaseFilterBackend
from url_filter.integrations.drf import DjangoFilterBackend
from rest_framework.exceptions import APIException

from .models import Viewpoint, Picture, Campaign


class BadFilter(APIException):
    status_code = 400
    default_detail = "Bad filter"
    default_code = "bad_filter"


class JsonFilterBackend(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        for key, field in settings.TROPP_SEARCHABLE_PROPERTIES.items():
            search_key = f'properties__{field["json_key"]}'
            if field["type"] == "many":
                search_item = request.GET.getlist(f"{search_key}")
            else:
                search_item = request.GET.get(search_key)
            if search_item:
                if field["type"] == "text":
                    # Full text search
                    queryset = queryset.annotate(
                        **{key: KeyTextTransform(field["json_key"], "properties")}
                    ).filter(**{f"{key}__icontains": search_item})
                else:
                    # Search on element
                    queryset = queryset.filter(
                        **{f"{search_key}__contains": search_item}
                    )
        return queryset

    def get_schema_fields(self, view):
        fields = []
        for key, field in settings.TROPP_SEARCHABLE_PROPERTIES.items():
            if field["type"] == "many":
                klass = coreschema.Array
            else:
                klass = coreschema.String
            description = f"{key.capitalize()} property ({field['type']})"
            fields.append(
                coreapi.Field(
                    name=f'properties__{field["json_key"]}',
                    required=False,
                    location="query",
                    schema=klass(
                        title=f"{key.capitalize()} property",
                        description=description,
                    ),
                )
            )
        return super().get_schema_fields(view)


class ViewpointFilterSet(FilterSet):
    id = filters.CharFilter(field_name="id", lookup_expr="exact")
    city = filters.CharFilter(field_name="city__label", lookup_expr="exact")
    themes = filters.CharFilter(field_name="themes__label", lookup_expr="exact")
    city_id = filters.CharFilter(field_name="city", lookup_expr="exact")
    themes_id = filters.CharFilter(field_name="themes", lookup_expr="exact")
    date_from = filters.DateFilter(field_name="pictures__date", lookup_expr="gte")
    date_to = filters.DateFilter(field_name="pictures__date", lookup_expr="lte")
    last_picture = filters.DateFilter(
        field_name="last_accepted_picture_date", lookup_expr="lte"
    )

    class Meta:
        model = Viewpoint
        fields = [
            "id",
            "city",
            "themes",
            "city_id",
            "themes_id",
            "date_from",
            "date_to",
            "active",
            "last_picture",
        ]


class PictureFilterSet(FilterSet):
    # owner_id = filters.IntegerFilter(field_name="owner", lookup_expr="exact")

    class Meta:
        model = Picture
        fields = ["owner"]


class CampaignFilterSet(FilterSet):
    date_from = filters.DateFilter(field_name="start_date", lookup_expr="gte")
    date_to = filters.DateFilter(field_name="start_date", lookup_expr="lte")

    class Meta:
        model = Campaign
        fields = [
            "date_from",
            "date_to",
        ]


class CampaignFilterBackend(BaseFilterBackend):
    """
    Filters for campaigns
    """

    def get_schema_fields(self, view):
        super().get_schema_fields(view)
        return [
            coreapi.Field(
                name="state",
                required=False,
                location="query",
                schema=coreschema.Boolean(
                    title="Campaign state",
                    description="'draft', 'started' or 'closed'",
                ),
            ),
            coreapi.Field(
                name="picture__state",
                required=False,
                location="query",
                schema=coreschema.Enum(
                    Picture.STATES,
                    description=str(pformat(Picture.STATES)),
                    title="Picture state",
                ),
            ),
        ]

    def filter_queryset(self, request, queryset, view):
        state = request.GET.get("state", None)
        if state is not None:

            if state not in [s[0] for s in Campaign.STATES]:
                raise BadFilter("Bad filter value for campaign state")

            queryset = queryset.filter(state=state)

        pictures_state = request.GET.get("pictures__state", None)
        if pictures_state is not None:

            if pictures_state not in [s[0] for s in Picture.STATES]:
                raise BadFilter("Bad filter value for pictures state")

            queryset = queryset.filter(pictures__state=pictures_state)

        return queryset


class SchemaAwareDjangoFilterBackend(DjangoFilterBackend):
    def get_schema_fields(self, view):
        """
        Get coreapi filter definitions

        Returns all schemas defined in filter_fields_schema attribute.
        """
        super().get_schema_fields(view)
        return getattr(view, "filter_fields_schema", [])



"""
Django settings for test_opp project.

Generated by 'django-admin startproject' using Django 2.2.3.

For more information on this file, see
https://docs.djangoproject.com/en/2.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.2/ref/settings/
"""

import os
import tempfile

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '+10auwvyy9--087ljr2o_-z^mg^@rx)*pe9--eikkn356awcna'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ['*', ]


# Application definition
INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.gis',
    'django.contrib.staticfiles',
    'versatileimagefield',
    'terra_accounts',
    'datastore',
    'geostore',
    'rest_framework',
    'rest_framework_gis',
    'terra_settings',
    'terra_opp',
]

INTERNAL_GEOMETRY_SRID = 4326

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'test_opp.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.dummy.DummyCache',
    }
}

MEDIA_ROOT = tempfile.gettempdir()

WSGI_APPLICATION = 'test_opp.wsgi.application'


# Database
# https://docs.djangoproject.com/en/2.2/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.contrib.gis.db.backends.postgis',
        'USER': 'travis_ci_test',
        'NAME': 'travis_ci_test',
        'PASSWORD': 'travis_ci_test',
        'HOST': os.getenv('POSTGRES_HOST', '127.0.0.1')
    }
}


# Password validation
# https://docs.djangoproject.com/en/2.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.2/howto/static-files/

STATIC_URL = '/static/'
MEDIA_URL = '/media/'
MEDIA_ROOT = os.path.join(BASE_DIR, 'test_opp', 'media')
SWAGGER_ENABLED = False

# set required DRF settings
REST_FRAMEWORK = {
    'TEST_REQUEST_DEFAULT_FORMAT': 'json',
    'DEFAULT_PAGINATION_CLASS': 'terra_settings.pagination.PagePagination',
    'PAGE_SIZE': 100,
}

AWS_DEFAULT_ACL = None

DEFAULT_FILE_STORAGE = 'django.core.files.storage.FileSystemStorage'

VERSATILEIMAGEFIELD_RENDITION_KEY_SETS = {
    'terra_opp': [
        ('original', 'url'),
        ('full', 'thumbnail__1500x1125'),
        ('list', 'thumbnail__300x225'),
        ('thumbnail', 'thumbnail__180x120'),
        ('thumbnail_vertical', 'thumbnail__120x180'),
    ]
}

AUTH_USER_MODEL = 'terra_accounts.TerraUser'
TROPP_OBSERVATORY_LAYER_PK = 1
TERRA_APPLIANCE_SETTINGS = {
    "disabled_modules": [],
}



import io
import operator
from functools import reduce

import coreapi
import coreschema

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.postgres.fields.jsonb import KeyTransform
from django.core.cache import cache
from django.db.models import Prefetch
from django.db import models
from django.template import loader
from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from django_filters.rest_framework import DjangoFilterBackend

from rest_framework import viewsets, renderers
from rest_framework.decorators import action
from rest_framework.filters import SearchFilter, OrderingFilter
from rest_framework.response import Response
from rest_framework.exceptions import APIException

from .filters import (
    CampaignFilterBackend,
    CampaignFilterSet,
    JsonFilterBackend,
    ViewpointFilterSet,
    SchemaAwareDjangoFilterBackend,
    PictureFilterSet,
)
from .models import Campaign, City, Picture, Theme, Viewpoint
from .pagination import RestPageNumberPagination
from .point_utilities import remove_point_thumbnail, update_point_properties
from .renderers import PdfRenderer, ZipRenderer, write_pdf
from . import permissions

from .serializers import (
    CampaignSerializer,
    RoCampaignSerializer,
    ListCampaignNestedSerializer,
    PictureSerializer,
    SimpleAuthenticatedViewpointSerializer,
    SimpleViewpointSerializer,
    ViewpointSerializerWithPicture,
    PhotographSerializer,
    CitySerializer,
    ThemeSerializer,
)


class CampaignNotFound(APIException):
    status_code = 400
    default_detail = "Can't found any valid campaign to add picture"
    default_code = "campaign_not_found"


class PictureAlreadyExists(APIException):
    status_code = 400
    default_detail = "Picture already exists for this viewpoint in that campaign"
    default_code = "picture_already_exists"


class ViewpointViewSet(viewsets.ModelViewSet):
    serializer_class = ViewpointSerializerWithPicture
    permission_classes = [
        permissions.ViewpointPermission,
    ]
    filter_backends = (
        SearchFilter,
        SchemaAwareDjangoFilterBackend,
        JsonFilterBackend,
        DjangoFilterBackend,
        OrderingFilter,
    )
    filterset_class = ViewpointFilterSet
    filter_fields_schema = [
        coreapi.Field(
            name="pictures__identifier",
            required=False,
            location="query",
            schema=coreschema.Integer(
                title="Picture identifier",
                description="Picture identifier to filter on",
            ),
        ),
        coreapi.Field(
            name="pictures__owner__uuid",
            required=False,
            location="query",
            schema=coreschema.Integer(
                title="Photographer uuid",
                description="Photographer uuid to filter on",
            ),
        ),
    ]
    filter_fields = ["pictures", "active"]
    search_fields = ("label", "id")
    pagination_class = RestPageNumberPagination
    template_name = "terra_opp/viewpoint_pdf.html"
    ordering_fields = ["id", "city", "active", "label"]
    ordering = ["-created_at"]

    def perform_create(self, serializer):
        serializer.save()
        update_point_properties(serializer.instance, self.request)

    def perform_update(self, serializer):
        serializer.save()
        update_point_properties(serializer.instance, self.request)

    def perform_destroy(self, instance):
        instance.point.delete()
        instance.delete()
        # FIXME This may be better if done directly in geostore
        # FIXME Try to be more precise and delete only the related feature's tile in cache
        cache.delete("tile_cache_*")  # delete all the cached tiles

    def get_queryset(self):
        # unauthenticated user not allowed to access archived viewpoint
        qs = Viewpoint.objects.with_accepted_pictures().filter(active=True)
        if self.request.user.is_authenticated:
            qs = Viewpoint.objects.all().distinct()

        pictures_qs = Picture.objects.order_by("-created_at")

        return (
            qs.select_related("point", "city")
            .prefetch_related(
                "pictures",
                "related",
                Prefetch("pictures", queryset=pictures_qs, to_attr="_ordered_pics"),
                "themes",
            )
            .annotate(
                last_accepted_picture_date=models.Max("pictures__date"),
            )
        )

    def get_serializer_class(self):
        if self.action == "list":
            if self.request.user.is_anonymous:
                return SimpleViewpointSerializer
            return SimpleAuthenticatedViewpointSerializer
        return ViewpointSerializerWithPicture

    @action(detail=False)
    def filters(self, request, *args, **kwargs):
        filter_values = {}
        for key, field in settings.TROPP_SEARCHABLE_PROPERTIES.items():
            data = None
            transform = KeyTransform(field["json_key"], "properties")
            queryset = (
                Viewpoint.objects.annotate(**{key: transform})
                .exclude(**{f"{key}__isnull": True})
                .values_list(key, flat=True)
            )
            if field["type"] == "single":
                # Dedupe and sort with SQL
                data = queryset.order_by(key).distinct(key)
            elif field["type"] == "many":
                # Dedupe and sort programmatically
                data = list(queryset)
                if data:
                    data = set(reduce(operator.concat, data))
                    data = sorted(data, key=str.lower)
            if data is not None:
                filter_values[key] = data

        filter_values["photographers"] = PhotographSerializer(
            get_user_model().objects.filter(pictures__isnull=False).distinct(),
            many=True,
        ).data

        # FIXME We may want to set all cities and themes as uniques directly in the model?
        filter_values["cities"] = (
            City.objects.exclude(label__isnull=True)
            .exclude(label__exact="")
            .distinct()
            .order_by(
                "label",
            )
            .values_list(
                "label",
                flat=True,
            )
        )
        filter_values["themes"] = (
            Theme.objects.exclude(label__isnull=True)
            .exclude(
                label__exact="",
            )
            .distinct()
            .order_by(
                "label",
            )
            .values_list(
                "label",
                flat=True,
            )
        )

        return Response(filter_values)

    @method_decorator(cache_page(60 * 5))
    @action(
        detail=True,
        methods=[
            "get",
        ],
        renderer_classes=[ZipRenderer],
        url_path="zip-pictures",
    )
    def zip_pictures(self, request, *args, **kwargs):
        qs = (
            self.get_object()
            .pictures.filter(
                state="accepted",
            )
            .only("file")
        )
        return Response([p.file for p in qs])

    @action(
        detail=True,
        renderer_classes=[renderers.TemplateHTMLRenderer],
    )
    def preview(self, request, *args, **kwargs):
        properties_set = settings.TROPP_VIEWPOINT_PROPERTIES_SET["pdf"]
        return Response(
            {
                "viewpoint": self.get_object(),
                "properties_set": properties_set,
            },
            template_name="terra_opp/viewpoint_pdf.html",
        )

    @method_decorator(cache_page(60 * 5))
    @action(
        detail=True,
        methods=[
            "get",
        ],
        renderer_classes=[PdfRenderer],
    )
    def pdf(self, request, *args, **kwargs):
        properties_set = settings.TROPP_VIEWPOINT_PROPERTIES_SET["pdf"]
        return Response(
            {
                "viewpoint": self.get_object(),
                "properties_set": properties_set,
            }
        )

    @action(detail=False, methods=["get"])
    def active(self, request, *args, **kwargs):
        qs = self.get_queryset().filter(active=True)
        qs_filtered = self.filter_queryset(qs)
        page = self.paginate_queryset(qs_filtered)

        if page is not None:
            serializer = SimpleViewpointSerializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = SimpleViewpointSerializer(qs_filtered, many=True)
        return Response(serializer.data)


class PictureViewSet(viewsets.ModelViewSet):
    queryset = Picture.objects.all()
    serializer_class = PictureSerializer
    permission_classes = [
        permissions.PicturePermission,
    ]
    filter_backends = (SearchFilter, DjangoFilterBackend, OrderingFilter)
    filterset_class = PictureFilterSet
    filter_fields = ["state", "active", "owner"]
    search_fields = ("viewpoint__label", "identifier")
    ordering_fields = ["state", "owner", "viewpoint", "identifier", "date"]
    ordering = ["-created_at"]
    pagination_class = RestPageNumberPagination

    def perform_create(self, serializer):
        if self.request.user.has_terra_perm("can_manage_pictures"):
            state = Picture.ACCEPTED
            owner = serializer.validated_data.get("owner")
            if not owner:
                owner = self.request.user
            serializer.save(state=state, owner=owner)
            update_point_properties(serializer.instance.viewpoint, self.request)

        elif self.request.user.has_terra_perm("can_add_pictures"):
            campaign = None
            viewpoint = serializer.validated_data.get("viewpoint")

            if "campaign" not in serializer.validated_data:
                # If no campaign specified we try to found one
                try:
                    campaign = Campaign.objects.get(
                        assignee=self.request.user,
                        state=Campaign.STARTED,
                        viewpoints=viewpoint,
                    )
                except Campaign.DoesNotExist:
                    raise CampaignNotFound()
            else:
                # Verify whether specified campaign exists
                campaign = serializer.validated_data.pop("campaign")
                try:
                    Campaign.objects.get(
                        id=campaign.id,
                        assignee=self.request.user,
                        state=Campaign.STARTED,
                        viewpoints=viewpoint,
                    )
                except Campaign.DoesNotExist:
                    raise CampaignNotFound()

            # Check whether the picture already exists for this viewpoint/campaign
            if Picture.objects.filter(viewpoint=viewpoint, campaign=campaign).exists():
                raise PictureAlreadyExists()

            serializer.save(owner=self.request.user, campaign=campaign)
            update_point_properties(serializer.instance.viewpoint, self.request)

    def perform_update(self, serializer):
        if self.request.user.has_terra_perm("can_manage_pictures"):
            serializer.save()
            update_point_properties(serializer.instance.viewpoint, self.request)

        elif self.request.user.has_terra_perm("can_add_pictures"):
            new_state = serializer.validated_data["state"]
            if new_state not in [
                Picture.DRAFT,
                Picture.SUBMITTED,
            ]:
                new_state = Picture.DRAFT
            # For self as user and draft state
            serializer.save(owner=self.request.user, state=new_state)
            update_point_properties(serializer.instance.viewpoint, self.request)

    def perform_destroy(self, instance):
        remove_point_thumbnail(instance, self.request)
        instance.delete()


class CampaignViewSet(viewsets.ModelViewSet):
    queryset = Campaign.objects.with_stats()
    permission_classes = [
        permissions.CampaignPermission,
    ]
    http_method_names = ["get", "post", "put", "delete", "options"]
    filterset_class = CampaignFilterSet
    filter_backends = (
        CampaignFilterBackend,
        DjangoFilterBackend,
        SearchFilter,
        OrderingFilter,
    )
    filter_fields = ["state", "assignee"]
    search_fields = ("label",)
    pagination_class = RestPageNumberPagination
    ordering_fields = ["label", "start_date", "assignee", "state"]
    ordering = ["-created_at"]

    def get_queryset(self):
        user = self.request.user
        qs = super().get_queryset()

        if self.action == "retrieve":
            qs = qs.prefetch_related("pictures")

        # Filter only on assigned campaigns for photographs
        if self.action == "list" and not user.has_terra_perm("can_manage_campaigns"):
            return qs.filter(
                assignee=user, state__in=[Campaign.STARTED, Campaign.CLOSED]
            )

        return qs

    def perform_create(self, serializer):
        serializer.save(owner=self.request.user)

    def get_serializer_class(self):
        if self.action == "list":
            return ListCampaignNestedSerializer

        # Serializer without stats for update
        if self.action in ["create", "update"]:
            return CampaignSerializer

        return RoCampaignSerializer

    @method_decorator(cache_page(60 * 5))
    @action(
        detail=True,
        renderer_classes=[ZipRenderer],
    )
    def all_sheets(self, request, *args, **kwargs):
        viewpoints = self.get_object().viewpoints.all()

        pdfs = []

        for viewpoint in viewpoints:
            template = loader.select_template(["terra_opp/viewpoint_pdf.html"])
            html = template.render(
                context={
                    "viewpoint": viewpoint,
                    "properties_set": settings.TROPP_VIEWPOINT_PROPERTIES_SET["pdf"],
                }
            )

            pdf_bytes = write_pdf(request, html)

            stream = io.BytesIO(pdf_bytes)
            stream.name = f"viewpoint_{viewpoint.pk}.pdf"

            pdfs.append(stream)

        return Response(pdfs)


class CityViewSet(viewsets.ModelViewSet):
    queryset = City.objects.all()
    http_method_names = ["get", "post", "put", "delete", "options"]
    permission_classes = [
        permissions.ViewpointPermission,
    ]
    filter_backends = (
        SearchFilter,
        OrderingFilter,
    )
    serializer_class = CitySerializer
    search_fields = ("label",)
    pagination_class = RestPageNumberPagination
    ordering_fields = ["label"]
    ordering = ["label"]


class ThemeViewSet(viewsets.ModelViewSet):
    queryset = Theme.objects.all()
    http_method_names = ["get", "post", "put", "delete", "options"]
    permission_classes = [
        permissions.ViewpointPermission,
    ]
    filter_backends = (
        SearchFilter,
        OrderingFilter,
        DjangoFilterBackend,
    )
    serializer_class = ThemeSerializer
    filter_fields = ["category"]
    search_fields = ("label", "category")
    pagination_class = RestPageNumberPagination
    ordering_fields = ["label", "category"]
    ordering = ["label"]
